"""
Pycroscopy's processing module

Submodules
----------

.. autosummary::
    :toctree: _autosummary

    atom_finding
    cluster
    decomposition
    feature_extraction
    fft
    giv_utils
    gmode_utils
    image_processing
    image_transformation
    proc_utils
    process
    svd_utils

"""
from . import fft
from . import gmode_utils
from . import proc_utils
from . import svd_utils
from .svd_utils import doSVD, rebuild_svd
from . import decomposition
from .decomposition import Decomposition
from . import cluster
from .cluster import Cluster
from . import image_processing
from .image_processing import ImageWindow
from . import giv_utils

def no_impl(*args,**kwargs):
    raise NotImplementedError("You need to install Multiprocess package (pip,github) to do a parallel Computation.\n"
                              "Switching to the serial version. ")

from .feature_extraction import FeatureExtractorParallel, FeatureExtractorSerial
from .image_transformation import geoTransformerParallel, geoTransformerSerial

FeatureExtractor = FeatureExtractorSerial
geoTransformer = geoTransformerSerial

try:
    import multiprocessing
except ImportError:
    FeatureExtractorParallel = no_impl
    geoTransformerParallel = no_impl
else:
    FeatureExtractor = FeatureExtractorParallel
    geoTransformer = geoTransformerParallel

__all__ = ['Cluster', 'Decomposition', 'ImageWindow', 'doSVD', 'fft', 'gmode_utils', 'proc_utils', 'svd_utils',
           'giv_utils', 'rebuild_svd']



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 15 11:10:37 2015

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import numpy as np

from .hdf_utils import getAuxData

__all__ = [
    'maxReadPixels', 'getActiveUDVSsteps', 'getDataIndicesForUDVSstep', 'getForExcitWfm',
    'getIndicesforPlotGroup', 'getSliceForExcWfm', 'generateTestSpectroscopicData', 'getSpecSliceForUDVSstep',
    'isSimpleDataset', 'reshapeToNsteps', 'reshapeToOneStep'
    ]


def maxReadPixels(max_memory, tot_pix, bins_per_step, bytes_per_bin=4):
    """
    Calculates the maximum number of pixels that can be loaded into the 
    specified memory size. This is particularly useful when applying a 
    (typically parallel) operation / processing on each pixel. 
    Example - Fitting response to a model.
    
    Parameters
    -------------
    max_memory : unsigned int
        Maximum memory (in bytes) that can be used. 
        For example 4 GB would be = 4*((2**10)**3) bytes
    tot_pix : unsigned int
        Total number of pixels in dataset
    bins_per_step : unsigned int
        Number of bins that will be read (can be portion of each pixel)
    bytes_per_bin : (Optional) unsigned int
        size of each bin - set to 4 bytes
    
    Returns
    -------
    max_pix : unsigned int
        Maximum number of pixels that will be loaded
    """
    # alternatively try .nbytes
    bytes_per_step = bins_per_step * bytes_per_bin
    max_pix = np.rint(max_memory / bytes_per_step)
    #print('Allowed to read {} of {} pixels'.format(max_pix,tot_pix))
    max_pix = max(1,min(tot_pix, max_pix))
    return np.uint(max_pix)

def getActiveUDVSsteps(h5_raw):
    """
    Returns all the active UDVS steps in the data
    
    Parameters
    ----------
    h5_raw : HDF5 dataset reference
        Reference to the raw data
        
    Returns
    -----------
    steps : 1D numpy array
        Active UDVS steps
    """
    udvs_step_vec = getAuxData(h5_raw, auxDataName =['UDVS_Indices'])[0].value
    return np.unique(udvs_step_vec)
    
def getSliceForExcWfm(h5_bin_wfm, excit_wfm):
    """
    Returns the indices that correspond to the given excitation waveform
    that can be used to slice the bin datasets
    * Developer note - Replace the first parameter with the Raw_Data dataset
    
    Parameters
    ----------------
    h5_bin_wfm : Reference to HDF5 dataset
        Bin Waveform Indices
    excit_wfm : integer
        excitation waveform / wave type
    
    Returns
    --------------
    slc : slice object
        Slice with the start and end indices
    """
    temp = np.where(h5_bin_wfm.value == excit_wfm)[0]
    return slice(temp[0],temp[-1]+1) # Need to add one additional index otherwise, the last index will be lost
    
def getDataIndicesForUDVSstep(h5_udvs_inds, udvs_step_index):
    """
    Returns the spectroscopic indices that correspond to the given udvs_step_index
    that can be used to slice the main data matrix.
    * Developer note - Replace the first parameter with the Raw_Data dataset
    
    Parameters
    -------------
    h5_udvs_inds : Reference to HDF5 dataset 
        UDVS_Indices dataset
    udvs_step_index : usigned int
        UDVS step index (base 0)
    
    Returns
    --------------
    ans : 1D numpy array 
        Spectroscopic indices
    """
    spec_ind_udvs_step_col = h5_udvs_inds[h5_udvs_inds.attrs.get('UDVS_Step')]
    return np.where(spec_ind_udvs_step_col == udvs_step_index)[0]
    
def getSpecSliceForUDVSstep(h5_udvs_inds, udvs_step_index):
    """
    Returns the spectroscopic indices that correspond to the given udvs_step_index
    that can be used to slice the main data matrix
    * Developer note - Replace the first parameter with the Raw_Data dataset
    
    Parameters
    -------------
    h5_udvs_inds : Reference to HDF5 dataset
        UDVS_Indices dataset
    udvs_step_index : unsigned int
        UDVS step index (base 0)
    
    Returns
    ----------
    slc : slice object
        Object containing the start and end indices
    """
    temp = np.where(h5_udvs_inds.value == udvs_step_index)[0]
    return slice(temp[0],temp[-1]+1) # Need to add one additional index otherwise, the last index will be lost
    

def getForExcitWfm(h5_main, h5_other, wave_type):
    """
    Slices the provided H5 dataset by the provided wave type. 
    Note that this is applicable to only certain H5 datasets such as the bin frequences, bin FFT etc.
    
    Parameters
    ----------
    h5_main : Reference to HDF5 dataset 
        Raw_Data dataset
    h5_other :Reference to HDF5 dataset 
        The dataset that needs to be sliced such as bin frequencies
    wave_type : unsigned int
        Excitation waveform type

    Returns
    ---------
    freq_vec : 1D numpy array
        data specific to specified excitation waveform
    """    
    h5_bin_wfm_type = getAuxData(h5_main, auxDataName=['Bin_Wfm_Type'])[0]
    inds = np.where(h5_bin_wfm_type.value == wave_type)[0]
    return h5_other[slice(inds[0],inds[-1]+1)]

    
def getIndicesforPlotGroup(h5_udvs_inds, ds_udvs, plt_grp_name):
    """
    For a provided plot group name in the udvs table, this function 
    returns the corresponding spectroscopic indices that can be used to index / slice the main data set
    and the data within the udvs table for the requested plot group
    * Developer note - Replace the first parameter with the Raw_Data dataset
        
    Parameters
    ------------
    h5_udvs_inds : Reference to HDF5 dataset
        containing the UDVS indices
    ds_udvs : Reference to HDF5 dataset 
        containing the UDVS table
    plt_grp_name : string 
        name of the plot group in the UDVS table
            
    Returns
    -----------
    step_bin_indices : 2D numpy array
        Indices arranged as [step, bin] in the spectroscopic_indices table
        This is useful for knowing the number of bins and steps in this plot group.
        We are allowed to assume that the number of bins does NOT change within the plot group
    oneD_indices : 1D numpy array
        spectroscopic indices corresponding to the requested plot group
    udvs_plt_grp_col : 1D numpy array
        data contained within the udvs table for the requested plot group        
    """
    
    # working on the UDVS table first:
    # getting the numpy array corresponding the requested plot group
    udvs_col_data = np.squeeze(ds_udvs[ds_udvs.attrs.get(plt_grp_name)])
    # All UDVS steps that are NOT part of the plot grop are empty cells in the table
    # and hence assume a nan value.
    # getting the udvs step indices that belong to this plot group:
    step_inds = np.where(np.isnan(udvs_col_data) == False)[0]    
    # Getting the values in that plot group that were non NAN
    udvs_plt_grp_col = udvs_col_data[step_inds]
    
    #---------------------------------
    
    # Now we use the udvs step indices calculated above to get 
    # the indices in the spectroscopic indices table
    spec_ind_udvs_step_col = h5_udvs_inds[h5_udvs_inds.attrs.get('UDVS_Step')]
    num_bins = len(np.where(spec_ind_udvs_step_col == step_inds[0])[0])
    # Stepehen says that we can assume that the number of bins will NOT change in a plot group
    step_bin_indices = np.zeros(shape=(len(step_inds),num_bins), dtype=int)

    for indx, step in enumerate(step_inds):
        step_bin_indices[indx,:] = np.where(spec_ind_udvs_step_col == step)[0]
    
    oneD_indices = step_bin_indices.reshape((step_bin_indices.shape[0]*step_bin_indices.shape[1]))
    return (step_bin_indices, oneD_indices, udvs_plt_grp_col)
    
def reshapeToOneStep(raw_mat, num_steps):
    """
    Reshapes provided data from (pos, step * bin) to (pos * step, bin). 
    This is useful when unraveling data for parallel processing.
    
    Parameters
    -------------
    raw_mat : 2D numpy array
        Data organized as (positions, step * bins)
    num_steps : unsigned int
        Number of spectroscopic steps per pixel (eg - UDVS steps)
        
    Returns
    --------------
    twoD : 2D numpy array
        Data rearranged as (positions * step, bin)
    """
    num_pos = raw_mat.shape[0]
    num_bins = int(raw_mat.shape[1]/num_steps)
    oneD = raw_mat
    oneD = oneD.reshape((num_bins * num_steps * num_pos))
    twoD = oneD.reshape((num_steps * num_pos, num_bins))
    return twoD
    
def reshapeToNsteps(raw_mat, num_steps):
    """
    Reshapes provided data from (positions * step, bin) to (positions, step * bin).  
    Use this to restructure data back to its original form after parallel computing 
    
    Parameters
    --------------
    raw_mat : 2D numpy array 
        Data organized as (positions * step, bin)
    num_steps : unsigned int
         Number of spectroscopic steps per pixel (eg - UDVS steps)
        
    Returns
    --------------- 
    twoD : 2D numpy array
        Data rearranged as (positions, step * bin)
    """
    num_bins = raw_mat.shape[1]
    num_pos = int(raw_mat.shape[0]/num_steps)
    oneD = raw_mat
    oneD = oneD.reshape(num_bins * num_steps * num_pos)
    twoD = oneD.reshape((num_pos, num_steps * num_bins)) 
    return twoD
    
def generateTestSpectroscopicData(num_bins=7, num_steps=3, num_pos=4):
    """
    Generates a (preferably small) test data set using the given parameters.
    Data is filled with indices (with base 1 for simplicity). 
    Use this for testing reshape operations etc.
    
    Parameters
    ----------
    num_bins : unsigned int (Optional. Default = 7)
        Number of bins
    num_steps : unsigned int (Optional. Default = 3)
        Number of spectroscopic steps  
    num_pos : unsigned int (Optional. Default = 4)
        Number of fictional positions
    
    Returns
    --------------
    full_data : 2D numpy array
        Data organized as [steps x bins, positions]
    """
    full_data = np.zeros((num_steps * num_bins, num_pos))
    for pos in range(num_pos):
        bin_count=0
        for step in range(num_steps):
            for bind in range(num_bins):
                full_data[bin_count, pos] = (pos+1)*100 + (step+1)*10 + (bind+1)
                bin_count += 1
    return full_data


def isSimpleDataset(h5_main, isBEPS=True):
    """
    This function figures out if a single number defines the bins for all UDVS steps
    In such cases (udvs_steps x bins, pos) can be reshaped to (bins, positions x steps)
    for (theoretically) faster computation, especially for large datasets

    Actually, things are a lot simpler. Only need to check if number of bins for all excitation waveforms are equal
    
    Parameters
    -------------
    h5_main : Reference to HDF5 dataset
        Raw_Data dataset
    isBEPS : Boolean (default = True)
        Whether or not this dataset is BEPS
        
    Returns
    ----------
    data_type : Boolean
        Whether or not this dataset can be unraveled / flattened
    """
    
    if isBEPS:
        if h5_main.parent.parent.attrs['VS_mode'] in ['DC modulation mode','AC modulation mode with time reversal','current mode','Relaxation']:
            # I am pretty sure that AC modulation also is simple
            return True
        else:
            # Could be user defined or some other kind I am not aware of
            # In many cases, some of these datasets could also potentially be simple datasets
            ds_udvs = getAuxData(h5_main, auxDataName=['UDVS'])[0]                
            excit_wfms = ds_udvs[ds_udvs.attrs.get('wave_mod')]
            wfm_types = np.unique(excit_wfms)
            if len(wfm_types) == 1:
                # BEPS with single excitation waveform
                print('Single BEPS excitation waveform')
                return True
            else:
                # Multiple waveform types here
                harm_types = np.unique(np.abs(wfm_types))
                if len(harm_types) != 1:
                    # eg - excitaiton waveforms 1, 2, 3 NOT -1, +1
                    return False
                # In this case a single excitation waveform with forward and reverse was used.
                h5_bin_wfm_type = getAuxData(h5_main, auxDataName=['Bin_Wfm_Type'])[0]
                # Now for each wfm type, count number of bins.
                wfm_bin_count = []
                for wfm in wfm_types:
                    wfm_bin_count.append(np.where(h5_bin_wfm_type.value == wfm)[0])
                wfm_lengths = np.unique(np.array(wfm_bin_count))
                if len(wfm_lengths) == 1:
                    # BEPS with multiple excitation waveforms but each excitation waveform has same number of bins
                    print('All BEPS excitation waves have same number of bins')
                    return True
            return False   
    else:
        # BE-Line
        return True


def isReshapable(h5_main, step_start_inds=None):
    """
    A BE dataset is said to be reshape-able if the number of bins per steps is constant. Even if the dataset contains
    multiple excitation waveforms (harmonics), We know that the measurement is always at the resonance peak, so the
    frequency vector should not change.

    Parameters
    ----------
    h5_main : h5py.Dataset object
        Reference to the main dataset
    step_start_inds : list or 1D array
        Indices that correspond to the start of each BE pulse / UDVS step

    Returns
    ---------
    reshapable : Boolean
        Whether or not the number of bins per step are constant in this dataset
    """
    if step_start_inds is None:
        h5_spec_inds = getAuxData(h5_main, auxDataName=['Spectroscopic_Indices'])[0]
        step_start_inds = np.where(h5_spec_inds[0] == 0)[0]
    # Adding the size of the main dataset as the last (virtual) step
    step_start_inds = np.hstack((step_start_inds, h5_main.shape[1]))
    num_bins = np.diff(step_start_inds)
    step_types = np.unique(num_bins)
    return len(step_types) == 1


# -*- coding: utf-8 -*-
"""
Created on Tue Oct  6 15:34:12 2015

@author: Numan Laanait -- nlaanait@gmail.com
"""

from __future__ import division, print_function, absolute_import
import warnings
import multiprocessing as mp

import h5py
import numpy as np
import skimage.feature


# TODO: Docstrings following numpy standard.

#### Functions
def pickle_keypoints(keypoints):
    """
    Function to pickle cv2.sift keypoint objects

    Parameters
    ----------
    keypoints
        keypoint object to be pickled

    Returns
    -------

    """
    kpArray = np.array([])
    for point in keypoints:
        kpArray = np.append(kpArray, [point.pt[1], point.pt[0]])
    kpArray = np.reshape(kpArray, (int(kpArray.size / 2), 2))
    return kpArray


# Class to do feature extraction. This is a wrapper on scikit-image and openCV feature extraction detectors.
# TODO: Add support for opencV or implement sift.
# TODO: Add io operations for extracted features.
# TODO: Memory checking, since some of the features are quite large.

class FeatureExtractorParallel(object):
    """
    This is an Object used to contain a data set and has methods to perform
    feature extraction on the data set that are detector based.
    Begin by loading a detector for features and a computer vision library.

    Parameters
    ----------
    detector_name : (string)
        name of detector.
    lib : (string)
        computer vision library to use (opencv or skimage)

        The following can be used for:
        lib = opencv: SIFT, ORB, SURF
        lib = skimage: ORB, BRIEF, CENSURE

    """

    def __init__(self, detector_name, lib):
        self.data = []
        self.lib = lib

        try:
            if self.lib == 'opencv':
                pass
                #                detector = cv2.__getattribute__(detector_name)
            elif self.lib == 'skimage':
                self.detector = skimage.feature.__getattribute__(detector_name)
        except AttributeError:
            print('Error: The Library does not contain the specified detector')

    def clearData(self):
        del self.data
        self.data = []

    def loadData(self, dataset):
        """
        This is a Method that loads h5 Dataset to be corrected.

        Parameters
        ----------
        dataset : h5py.Dataset
            Dataset to be corrected
        """
        if not isinstance(dataset, h5py.Dataset):
            warnings.warn('Error: Data must be an h5 Dataset object')
        else:
            self.data = dataset
            dim = int(np.sqrt(self.data.shape[-1]))
            self.data = self.data.reshape(-1, dim, dim)

    def getData(self):
        """
        This is a Method that returns the loaded h5 Dataset.

        """
        return self.data

    def getFeatures(self, **kwargs):
        """
        This is a Method that returns features (keypoints and descriptors)
        that are obtained by using the FeatureExtractor.Detector object.

        Parameters
        ----------
        processors : int, optional
                    Number of processors to use, default = 1.
        mask : boolean, optional, default False.
            Whether to use

        Returns
        -------
        keypts :
            keypoints
        descs :
            descriptors

        """
        detector = self.detector
        dset = self.data
        lib = self.lib
        processes = kwargs.get('processors', 1)
        mask = kwargs.get('mask', False)
        origin = kwargs.get('origin', [0, 0])
        winSize = kwargs.get('window_size', 0)

        if mask:
            def mask_func(x, winSize):
                x[origin[0] - winSize / 2: origin[0] + winSize / 2,
                origin[1] - winSize / 2: origin[1] + winSize / 2] = 2
                x = x - 1
                return x

            mask_ind = np.mask_indices(dset.shape[-1], mask_func, winSize)
            self.data = np.array([imp[mask_ind].reshape(winSize, winSize) for imp in dset])

        # detect and compute keypoints
        def detect(image):
            if lib == 'opencv':
                image = (image - image.mean()) / image.std()
                image = image.astype('uint8')
                k_obj, d_obj = detector.detectAndCompute(image, None)
                keypts, descs = pickle_keypoints(k_obj), pickle_keypoints(d_obj)

            elif lib == 'skimage':
                imp = (image - image.mean()) / np.std(image)
                imp[imp < 0] = 0
                imp.astype('float32')
                detector.detect_and_extract(imp)
                keypts, descs = detector.keypoints, detector.descriptors

            return keypts, descs

        # start pool of workers
        print('launching %i kernels...' % (processes))
        pool = mp.Pool(processes)
        tasks = [(imp) for imp in self.data]
        chunk = int(self.data.shape[0] / processes)
        jobs = pool.imap(detect, tasks, chunksize=chunk)

        # get keypoints and descriptors
        results = []
        print('Extracting features...')
        try:
            for j in jobs:
                results.append(j)
        except ValueError:
            warnings.warn('ValueError something about 2d-image. Probably some of the detector input params are wrong.')

        keypts = [itm[0].astype('int') for itm in results]
        desc = [itm[1] for itm in results]

        # close the pool
        print('Closing down the kernels... \n')
        pool.close()

        return keypts, desc


class FeatureExtractorSerial(object):
    """
    This is an Object used to contain a data set and has methods to perform
    feature extraction on the data set that are detector based.
    Begin by loading a detector for features and a computer vision library.

    Parameters
    ----------
    detector_name : (string)
        name of detector.
    lib : (string)
        computer vision library to use (opencv or skimage)

        The following can be used for:
        lib = opencv: SIFT, ORB, SURF
        lib = skimage: ORB, BRIEF, CENSURE

    """

    def __init__(self, detector_name, lib):
        self.data = []
        self.lib = lib

        try:
            if self.lib == 'opencv':
                pass
                #                detector = cv2.__getattribute__(detector_name)
            elif self.lib == 'skimage':
                self.detector = skimage.feature.__getattribute__(detector_name)
        except AttributeError:
            print('Error: The Library does not contain the specified detector')

    def clearData(self):
        del self.data
        self.data = []

    def loadData(self, dataset):
        """
        This is a Method that loads h5 Dataset to be corrected.

        Parameters
        ----------
        dataset : h5py.Dataset

        """
        if not isinstance(dataset, h5py.Dataset):
            warnings.warn('Error: Data must be an h5 Dataset object')
        else:
            self.data = dataset
            dim = int(np.sqrt(self.data.shape[-1]))
            self.data = self.data.reshape(-1, dim, dim)

    def getData(self):
        """
        This is a Method that returns the loaded h5 Dataset.
        """
        return self.data

    def getFeatures(self, **kwargs):
        """
        This is a Method that returns features (keypoints and descriptors)
        that are obtained by using the FeatureExtractor.Detector object.

        Parameters
        ----------
        mask : boolean, optional
            Whether to use, default False.

        Returns
        -------
        keypts :
            descriptors
        descs :
            keypoints

        """
        detector = self.detector
        dset = self.data
        lib = self.lib
        mask = kwargs.get('mask', False)
        origin = kwargs.get('origin', [0, 0])
        winSize = kwargs.get('window_size', 0)

        if mask:
            def mask_func(x, winSize):
                x[origin[0] - winSize / 2: origin[0] + winSize / 2,
                origin[1] - winSize / 2: origin[1] + winSize / 2] = 2
                x = x - 1
                return x

            mask_ind = np.mask_indices(dset.shape[-1], mask_func, winSize)
            self.data = np.array([imp[mask_ind].reshape(winSize, winSize) for imp in dset])

        # detect and compute keypoints
        def detect(image):
            if lib == 'opencv':
                image = (image - image.mean()) / image.std()
                image = image.astype('uint8')
                k_obj, d_obj = detector.detectAndCompute(image, None)
                keypts, descs = pickle_keypoints(k_obj), pickle_keypoints(d_obj)

            elif lib == 'skimage':
                imp = (image - image.mean()) / np.std(image)
                imp[imp < 0] = 0
                imp.astype('float32')
                detector.detect_and_extract(imp)
                keypts, descs = detector.keypoints, detector.descriptors

            return keypts, descs

        # start pool of workers
        results = [detect(imp) for imp in self.data]

        # get keypoints and descriptors
        keypts = [itm[0].astype('int') for itm in results]
        desc = [itm[1] for itm in results]

        return keypts, desc



# -*- coding: utf-8 -*-
"""
Created on Tue Jan 05 07:55:56 2016

@author: Suhas Somnath, Chris Smith
"""
from __future__ import division, print_function, absolute_import
import numpy as np
import sklearn.cluster as cls
from scipy.cluster.hierarchy import linkage
from scipy.spatial.distance import pdist
from multiprocessing import cpu_count
from ..io.hdf_utils import getH5DsetRefs, checkAndLinkAncillary, copy_main_attributes, checkIfMain
from ..io.io_hdf5 import ioHDF5
from ..io.io_utils import check_dtype, transformToTargetType
from ..io.microdata import MicroDataGroup, MicroDataset


class Cluster(object):
    """
    Pycroscopy wrapper around the sklearn.cluster classes.
    """

    def __init__(self, h5_main, method_name, num_comps=None, *args, **kwargs):
        """
        Constructs the Cluster object

        Parameters
        ------------
        h5_main : HDF5 dataset object
            Main dataset with ancillary spectroscopic, position indices and values datasets
        method_name : string / unicode
            Name of the sklearn.cluster estimator
        num_comps : (optional) unsigned int
            Number of features / spectroscopic indices to be used to cluster the data. Default = all
        *args and **kwargs : arguments to be passed to the estimator
        """

        allowed_methods = ['AgglomerativeClustering', 'Birch', 'KMeans',
                           'MiniBatchKMeans', 'SpectralClustering']
        
        # check if h5_main is a valid object - is it a hub?
        if not checkIfMain(h5_main):
            raise TypeError('Supplied dataset is not a pycroscopy main dataset')

        if method_name not in allowed_methods:
            raise TypeError('Cannot work with {} just yet'.format(method_name))

        self.h5_main = h5_main

        '''
        If n_jobs is not provided, set to n_cores-2
        '''
        kwargs.update({'n_jobs': kwargs.pop('n_jobs', max(1, cpu_count() - 2))})

        # Instantiate the clustering object
        self.estimator = cls.__dict__[method_name].__call__(*args, **kwargs)
        self.method_name = method_name

        comp_slice, num_comps = self._get_component_slice(num_comps)

        self.num_comps = num_comps
        self.data_slice = (slice(None), comp_slice)

        # figure out the operation that needs need to be performed to convert to real scalar
        retval = check_dtype(h5_main)
        self.data_transform_func, self.data_is_complex, self.data_is_compound, \
        self.data_n_features, self.data_n_samples, self.data_type_mult = retval

    def do_cluster(self, rearrange_clusters=True):
        """
        Clusters the hdf5 dataset, calculates mean response for each cluster, and writes the labels and mean response
        back to the h5 file

        Parameters
        ----------
        rearrange_clusters : (Optional) Boolean. Default = True
            Whether or not the clusters should be re-ordered by relative distances between the mean response

        Returns
        --------
        h5_group : HDF5 Group reference
            Reference to the group that contains the clustering results
        """
        self._fit()
        new_mean_response = self._get_mean_response(self.results.labels_)
        new_labels = self.results.labels_
        if rearrange_clusters:
            new_labels, new_mean_response = reorder_clusters(self.results.labels_, new_mean_response)
        return self._write_to_hdf5(new_labels, new_mean_response)

    def _fit(self):
        """
        Fits the provided dataset

        Returns
        ------
        None
        """
        print('Performing clustering on {}.'.format(self.h5_main.name))
        # perform fit on the real dataset
        self.results = self.estimator.fit(self.data_transform_func(self.h5_main[self.data_slice]))

    def _get_mean_response(self, labels):
        """
        Gets the mean response for each cluster

        Parameters
        -------------
        labels : 1D unsigned int array
            Array of cluster labels as obtained from the fit

        Returns
        ---------
        mean_resp : 2D numpy array
            Array of the mean response for each cluster arranged as [cluster number, response]
        """
        print('Calculated the Mean Response of each cluster.')
        num_clusts = len(np.unique(labels))
        mean_resp = np.zeros(shape=(num_clusts, self.num_comps), dtype=self.h5_main.dtype)
        for clust_ind in range(num_clusts):
            # get all pixels with this label
            targ_pos = np.argwhere(labels == clust_ind)
            # slice to get the responses for all these pixels, ensure that it's 2d
            data_chunk = np.atleast_2d(self.h5_main[:, self.data_slice[1]][targ_pos, :])
            # transform to real from whatever type it was
            avg_data = np.mean(self.data_transform_func(data_chunk), axis=0, keepdims=True)
            # transform back to the source data type and insert into the mean response
            mean_resp[clust_ind] = transformToTargetType(avg_data, self.h5_main.dtype)
        return mean_resp

    def _get_component_slice(self, components):
        """
        Check the components object to determine how to use it to slice the dataset

        Parameters
        ----------
        components : {int, iterable of ints, slice, or None}
            Input Options
            integer: Components less than the input will be kept
            length 2 iterable of integers: Integers define start and stop of component slice to retain
            other iterable of integers or slice: Selection of component indices to retain
            None: All components will be used
        Returns
        -------
        comp_slice : slice or numpy array of uints
            Slice or array specifying which components should be kept
        """

        comp_slice = slice(None)

        if components is None:
            num_comps = self.h5_main.shape[1]
            comp_slice = slice(0, num_comps)
        elif isinstance(components, int):
            # Component is integer
            num_comps = int(np.min([components, self.h5_main.shape[1]]))
            comp_slice = slice(0, num_comps)
        elif hasattr(components, '__iter__') and not isinstance(components, dict):
            # Component is array, list, or tuple
            if len(components) == 2:
                # If only 2 numbers are given, use them as the start and stop of a slice
                comp_slice = slice(int(components[0]), int(components[1]))
                num_comps = abs(comp_slice.stop-comp_slice.start)
            else:
                #Convert components to an unsigned integer array
                comp_slice = np.uint(components)
                num_comps = len(comp_slice)
        elif isinstance(components, slice):
            # Components is already a slice
            comp_slice = components
            num_comps = abs(comp_slice.stop-comp_slice.start)
        elif components is not None:
            raise TypeError('Unsupported component type supplied to clean_and_build.  Allowed types are integer, numpy array, list, tuple, and slice.')

        return comp_slice, num_comps

    def _write_to_hdf5(self, labels, mean_response):
        """
        Writes the labels and mean response to the h5 file

        Parameters
        ------------
        labels : 1D unsigned int array
            Array of cluster labels as obtained from the fit
        mean_response : 2D numpy array
            Array of the mean response for each cluster arranged as [cluster number, response]

        Returns
        ---------
        h5_labels : HDF5 Group reference
            Reference to the group that contains the clustering results
        """
        print('Writing clustering results to file.')
        num_clusters = mean_response.shape[0]
        ds_label_mat = MicroDataset('Labels', np.uint32(labels.reshape([-1, 1])), dtype=np.uint32)
        clust_ind_mat = np.transpose(np.atleast_2d(np.arange(num_clusters)))

        ds_cluster_inds = MicroDataset('Cluster_Indices', np.uint32(clust_ind_mat))
        ds_cluster_vals = MicroDataset('Cluster_Values', np.float32(clust_ind_mat))
        ds_cluster_centroids = MicroDataset('Mean_Response', mean_response, dtype=mean_response.dtype)
        ds_label_inds = MicroDataset('Label_Spectroscopic_Indices', np.atleast_2d([0]), dtype=np.uint32)
        ds_label_vals = MicroDataset('Label_Spectroscopic_Values', np.atleast_2d([0]), dtype=np.float32)

        # write the labels and the mean response to h5
        clust_slices = {'Cluster': (slice(None), slice(0, 1))}
        ds_cluster_inds.attrs['labels'] = clust_slices
        ds_cluster_inds.attrs['units'] = ['']
        ds_cluster_vals.attrs['labels'] = clust_slices
        ds_cluster_vals.attrs['units'] = ['']

        cluster_grp = MicroDataGroup(self.h5_main.name.split('/')[-1] + '-Cluster_', self.h5_main.parent.name[1:])
        cluster_grp.addChildren([ds_label_mat, ds_cluster_centroids, ds_cluster_inds, ds_cluster_vals, ds_label_inds,
                                 ds_label_vals])

        cluster_grp.attrs['num_clusters'] = num_clusters
        cluster_grp.attrs['num_samples'] = self.h5_main.shape[0]
        cluster_grp.attrs['cluster_algorithm'] = self.method_name

        h5_spec_inds = self.h5_main.file[self.h5_main.attrs['Spectroscopic_Indices']]
        h5_spec_vals = self.h5_main.file[self.h5_main.attrs['Spectroscopic_Values']]

        '''
        Setup the Spectroscopic Indices and Values for the Mean Response if we didn't use all components
        '''
        if self.num_comps != self.h5_main.shape[1]:
            ds_centroid_indices = MicroDataset('Mean_Response_Indices', np.arange(self.num_comps, dtype=np.uint32))

            if isinstance(self.data_slice[1], np.ndarray):
                centroid_vals_mat = h5_spec_vals[self.data_slice[1].tolist()]
                cluster_grp.attrs['components_used'] = self.data_slice[1].tolist()

            else:
                centroid_vals_mat = h5_spec_vals[self.data_slice[1]]

                cluster_grp.attrs['components_used'] = range(self.data_slice[1].start,
                                                             self.data_slice[1].stop)[self.data_slice[1]]

            ds_centroid_values = MicroDataset('Mean_Response_Values', centroid_vals_mat)

            cluster_grp.addChildren([ds_centroid_indices, ds_centroid_values])

        else:
            cluster_grp.attrs['components_used'] = 'all'

        '''
        Get the parameters of the estimator used and write them
        as attributes of the group
        '''
        for parm in self.estimator.get_params().keys():
            cluster_grp.attrs[parm] = self.estimator.get_params()[parm]

        hdf = ioHDF5(self.h5_main.file)
        h5_clust_refs = hdf.writeData(cluster_grp)

        h5_labels = getH5DsetRefs(['Labels'], h5_clust_refs)[0]
        h5_centroids = getH5DsetRefs(['Mean_Response'], h5_clust_refs)[0]
        h5_clust_inds = getH5DsetRefs(['Cluster_Indices'], h5_clust_refs)[0]
        h5_clust_vals = getH5DsetRefs(['Cluster_Values'], h5_clust_refs)[0]
        h5_label_inds = getH5DsetRefs(['Label_Spectroscopic_Indices'], h5_clust_refs)[0]
        h5_label_vals = getH5DsetRefs(['Label_Spectroscopic_Values'], h5_clust_refs)[0]

        copy_main_attributes(self.h5_main, h5_centroids)

        if isinstance(self.data_slice[1], np.ndarray):
            h5_mean_resp_inds = getH5DsetRefs(['Mean_Response_Indices'], h5_clust_refs)[0]
            h5_mean_resp_vals = getH5DsetRefs(['Mean_Response_Values'], h5_clust_refs)[0]
        else:
            h5_mean_resp_inds = h5_spec_inds
            h5_mean_resp_vals = h5_spec_vals

        # h5_label_inds.attrs['labels'] = ''
        # h5_label_inds.attrs['units'] = ''
        # h5_label_vals.attrs['labels'] = ''
        # h5_label_vals.attrs['units'] = ''

        checkAndLinkAncillary(h5_labels,
                              ['Position_Indices', 'Position_Values'],
                              h5_main=self.h5_main)
        checkAndLinkAncillary(h5_labels,
                              ['Spectroscopic_Indices', 'Spectroscopic_Values'],
                              anc_refs=[h5_label_inds, h5_label_vals])

        checkAndLinkAncillary(h5_centroids,
                              ['Spectroscopic_Indices', 'Spectroscopic_Values'],
                              anc_refs=[h5_mean_resp_inds, h5_mean_resp_vals])

        checkAndLinkAncillary(h5_centroids,
                              ['Position_Indices', 'Position_Values'],
                              anc_refs=[h5_clust_inds, h5_clust_vals])

        # return the h5 group object
        return h5_labels.parent


def reorder_clusters(labels, mean_response):
    """
    Reorders clusters by the distances between the clusters

    Parameters
    ----------
    labels : 1D unsigned int numpy array
        Labels for the clusters
    mean_response : 2D numpy array
        Mean response of each cluster arranged as [cluster , features]

    Returns
    -------
    new_labels : 1D unsigned int numpy array
        Labels for the clusters arranged by distances
    new_mean_response : 2D numpy array
        Mean response of each cluster arranged as [cluster , features]
    """

    num_clusters = mean_response.shape[0]
    # Get the distance between cluster means
    distance_mat = pdist(mean_response)
    # get hierarchical pairings of clusters
    linkage_pairing = linkage(distance_mat, 'weighted')

    # get the new order - this has been checked to be OK
    new_cluster_order = []
    for row in range(linkage_pairing.shape[0]):
        for col in range(2):
            if linkage_pairing[row, col] < num_clusters:
                new_cluster_order.append(int(linkage_pairing[row, col]))

    # Now that we know the order, rearrange the clusters and labels:
    new_labels = np.zeros(shape=labels.shape, dtype=labels.dtype)
    new_mean_response = np.zeros(shape=mean_response.shape, dtype=mean_response.dtype)

    # Reorder clusters
    for old_clust_ind, new_clust_ind in enumerate(new_cluster_order):
        new_labels[np.where(labels == new_clust_ind)[0]] = old_clust_ind
        new_mean_response[old_clust_ind] = mean_response[new_clust_ind]

    return new_labels, new_mean_response



# -*- coding: utf-8 -*-
"""
Created on Tue Nov  3 15:07:16 2015

@author: Suhas Somnath

"""

from __future__ import division, print_function, absolute_import, unicode_literals

from os import path, listdir, remove
from warnings import warn

import numpy as np
import xlrd as xlreader  # To read the UDVS spreadsheet
from scipy.io.matlab import loadmat  # To load parameters stored in Matlab .mat file

from .df_utils.be_utils import trimUDVS, getSpectroscopicParmLabel, parmsToDict, generatePlotGroups, normalizeBEresponse, \
    createSpecVals, nf32
from .translator import Translator
from .utils import make_position_mat, generate_dummy_main_parms
from ..hdf_utils import getH5DsetRefs, linkRefs, calc_chunks
from ..io_hdf5 import ioHDF5
from ..microdata import MicroDataGroup, MicroDataset


class BEPSndfTranslator(Translator):
    """
    Translates Band Excitation Polarization Switching (BEPS) datasets from .dat
    files to .h5

    """       
        
    def translate(self, data_filepath, show_plots=True, save_plots=True, do_histogram=False, debug=False):
        """
        The main function that translates the provided file into a .h5 file
        
        Parameters
        ----------------
        data_filepath : String / unicode
            Absolute path of the data file (.dat)
        show_plots : Boolean (Optional. Default is True)
            Whether or not to show plots
        save_plots : Boolean (Optional. Default is True)
            Whether or not to save the generated plots
        do_histogram : Boolean (Optional. Default is False)
            Whether or not to generate and save 2D histograms of the raw data
        debug : Boolean (Optional. default is false)
            Whether or not to print log statements
            
        Returns
        --------------
        h5_path : String / unicode
            Absolute path of the generated .h5 file

        """
        ## Read the parameter files
        if debug:
            print('BEndfTranslator: Getting file paths')

        parm_filepath, udvs_filepath, parms_mat_path = self._parse_file_path(data_filepath)
        if debug:
            print('BEndfTranslator: Reading Parms text file')

        isBEPS, self.parm_dict = parmsToDict(parm_filepath)
        self.parm_dict['data_type'] = 'BEPSData'
        if not isBEPS:
            warn('This is NOT a BEPS new-data-format dataset!')
            return None
        
        """ Find out if this is a custom experiment and whether in and out of field were acquired
        For a standard experiment where only in / out field is acquired, zeros are stored
        even for those UDVS steps without band excitation"""
        self.field_mode = self.parm_dict['VS_measure_in_field_loops']
        expt_type = self.parm_dict['VS_mode']
        self.spec_label = getSpectroscopicParmLabel(expt_type)
        std_expt = expt_type in ['DC modulation mode', 'current mode']
        self.halve_udvs_steps = False
        ignored_plt_grps = []
        if std_expt and self.field_mode != 'in and out-of-field':
            self.halve_udvs_steps = True
            if self.field_mode == 'out-of-field':
                ignored_plt_grps = ['in-field']
            else:
                ignored_plt_grps = ['out-of-field']

        h5_path = path.join(self.folder_path, self.basename+'.h5')
        if path.exists(h5_path):
            remove(h5_path)
        
        if debug:
            print('BEndfTranslator: Preparing to read parms.mat file')
        self.BE_wave, self.BE_wave_rev, self.BE_bin_inds = self.__get_excit_wfm(parms_mat_path)
        
        if debug:
            print('BEndfTranslator: About to read UDVS file')
        
        self.udvs_labs, self.udvs_units, self.udvs_mat = self.__read_udvs_table(udvs_filepath)
        # Remove the unused plot group columns before proceeding:
        self.udvs_mat, self.udvs_labs, self.udvs_units = trimUDVS(self.udvs_mat, self.udvs_labs, self.udvs_units,
                                                                  ignored_plt_grps)
        if debug: print('BEndfTranslator: Read UDVS file')
            
        self.num_udvs_steps = self.udvs_mat.shape[0]
        # This is absolutely crucial for reconstructing the data chronologically
        self.excit_type_vec = (self.udvs_mat[:, 4]).astype(int)
        
        # First figure out how many waveforms are present in the data from the UDVS
        unique_waves = self.__get_unique_wave_types(self.excit_type_vec) 
        self.__unique_waves__ = unique_waves
        self.__num_wave_types__ = len(unique_waves)
        # print self.__num_wave_types__, 'different excitation waveforms in this experiment'
        
        if debug: print('BEndfTranslator: Preparing to set up parsers')

        # Preparing objects to parse the file(s)
        parsers = self.__assemble_parsers()        
        
        # Gathering some basic details before parsing the files:
        self.max_pixels = parsers[0].get_num_pixels()
        s_pixels = np.array(parsers[0].get_spatial_pixels())
        self.pos_labels = ['Laser Spot', 'Z', 'Y', 'X']
        self.pos_labels = [self.pos_labels[i] for i in np.where(s_pixels > 1)[0]]
        self.pos_mat = make_position_mat(s_pixels)
        self.pos_units = ['um' for _ in range(len(self.pos_labels))]
#         self.pos_mat = np.int32(self.pos_mat)
        
        # Helping Eric out a bit. Remove this section at a later time:
        main_parms = generate_dummy_main_parms()
        # main_parms['grid_size_x'] = self.parm_dict['grid_num_cols']
        # main_parms['grid_size_y'] = self.parm_dict['grid_num_rows']
        main_parms['grid_size_x'] = self.parm_dict['grid_num_rows']
        main_parms['grid_size_y'] = self.parm_dict['grid_num_cols']
        main_parms['experiment_date'] = self.parm_dict['File_date_and_time']
        # assuming that the experiment was completed:        
        main_parms['current_position_x'] = self.parm_dict['grid_num_rows']-1
        main_parms['current_position_y'] = self.parm_dict['grid_num_cols']-1
        main_parms['data_type'] = 'BEPSData'
        main_parms['translator'] = 'NDF'
        
        # Writing only the root now:
        spm_data = MicroDataGroup('')
        spm_data.attrs = main_parms
        self.hdf = ioHDF5(h5_path)
#         self.hdf.clear()
        
        #cacheSettings = self.hdf.file.id.get_access_plist().get_cache()
        #print 'H5 cache settings: Metadata Objects {}    Data Chunks {}    Raw Data Size(kB) {}'.format(cacheSettings[0],cacheSettings[1],cacheSettings[2]/1024)
        
        self.hdf.writeData(spm_data)
        
        ########################################################
        # Reading and parsing the .dat file(s) 

        self._read_data(parsers, unique_waves, show_plots, save_plots, do_histogram)
        
        self.hdf.close()
            
        return h5_path

    def _read_data(self, parsers, unique_waves, show_plots, save_plots, do_histogram):
        """
        Loops over all pixels and reads the data into the HDF5 file.

        Parameters
        ----------
        parsers : list of BEPSndfPixel
            List of parser object that will read the pixel data from the files
        unique_waves : numpy.ndarray of int
            Array denoting the unique waveforms in the experiment
        show_plots : Boolean, optional
            Should generated plots be shown during translation.  Default True
        save_plots : Boolean, optional
            Should generated plots be saved to disk during translation.  Default True
        do_histogram : Boolean, optional
            Should histograms be generated for the different plot groups.  Default False

        Returns
        -------
        None

        """
        print('Reading data file(s)')
        self.dset_index = 0
        self.ds_pixel_start_indx = 0
        for pixel_ind in range(self.max_pixels):

            if (100.0 * (pixel_ind + 1) / self.max_pixels) % 10 == 0:
                print('{} % complete'.format(int(100 * (pixel_ind + 1) / self.max_pixels)))

            # First read the next pixel from all parsers:
            current_pixels = {}
            for prsr in parsers:
                wave_type = prsr.get_wave_type()
                if self.parm_dict['VS_mode'] == 'AC modulation mode with time reversal' and \
                                self.BE_bin_inds is not None:
                    if np.sign(wave_type) == -1:
                        bin_fft = self.BE_wave[self.BE_bin_inds]
                    elif np.sign(wave_type) == 1:
                        bin_fft = self.BE_wave_rev[self.BE_bin_inds]
                else:
                    bin_fft = None

                current_pixels[wave_type] = prsr.read_pixel(bin_fft)

            if pixel_ind == 0:
                h5_refs = self.__initialize_meas_group(self.max_pixels, current_pixels)
                prev_pixels = current_pixels  # This is here only to avoid annoying warnings.
            else:
                if current_pixels[unique_waves[0]].is_different_from(prev_pixels[unique_waves[0]]):
                    # Some parameter has changed. Write current group and make new group
                    self.__close_meas_group(h5_refs, show_plots, save_plots, do_histogram)
                    self.ds_pixel_start_indx = pixel_ind
                    h5_refs = self.__initialize_meas_group(self.max_pixels - pixel_ind, current_pixels)

            # print('reading Pixel {} of {}'.format(pixel_ind,self.max_pixels))
            self.__append_pixel_data(current_pixels)

            prev_pixels = current_pixels
        self.__close_meas_group(h5_refs, show_plots, save_plots, do_histogram)

    ###################################################################################################
        
    def __close_meas_group(self, h5_refs, show_plots, save_plots, do_histogram):
        """
        Performs following operations : 
            * Updates the number of pixels attribute in the measurement group
            * Writes Noise floor axis labels as region references
            * Writes position values and indices along with region references
            * Links all ancilliary datasets to the main data set
            * Writes the spatiall averaged plot data
        
        Parameters
        ----------
        h5_refs : list of HDF references
            References to the written datasets
        show_plots : Boolean 
            Whether or not to show plots
        save_plots : Boolean
            Whether or not to save the generated plots
        do_histogram : Boolean
            Whether or not to generate and save 2D histograms of the raw data
            
        Returns
        -------
        None

        """
        # Update the number of pixels in the attributes
        meas_grp = self.ds_main.parent
        meas_grp.attrs['num_pix'] = self.ds_pixel_index
        
        # Write position specific datasets now that the dataset is complete
        pos_slice_dict = dict()
        for spat_ind, spat_dim in enumerate(self.pos_labels):
            pos_slice_dict[spat_dim] = (slice(None), slice(spat_ind, spat_ind+1))
        ds_pos_ind = MicroDataset('Position_Indices', self.pos_mat[self.ds_pixel_start_indx:
                                  self.ds_pixel_start_indx + self.ds_pixel_index, :], dtype=np.uint)
        ds_pos_ind.attrs['labels'] = pos_slice_dict
        ds_pos_ind.attrs['units'] = self.pos_units

        self.pos_vals_list = np.array(self.pos_vals_list)
        # Ensuring that the X and Y values vary from 0 to N instead of -0.5 N to + 0.5 N
        for col_ind in range(2):
            min_val = np.min(self.pos_vals_list[:, col_ind])
            self.pos_vals_list[:, col_ind] -= min_val
            self.pos_vals_list[:, col_ind] *= 1E+6  # convert to microns

        if np.max(self.pos_vals_list[:, 2]) > 1E-3:
            # Setpoint spectroscopy
            if 'Z' in self.pos_labels:
                dim_ind = self.pos_labels.index('Z')
                # TODO: Find a way to correct the labels
                # self.pos_labels[dim_ind] = 'Setpoint'
                self.pos_units[dim_ind] = 'defl V'
        else:
            # Z spectroscopy
            self.pos_vals_list[:, 2] *= 1E+6  # convert to microns

        pos_val_mat = np.float32(self.pos_mat[self.ds_pixel_start_indx:
                                              self.ds_pixel_start_indx + self.ds_pixel_index, :])

        for col_ind, targ_dim_name in enumerate(['X', 'Y', 'Z']):
            if targ_dim_name in self.pos_labels:
                dim_ind = self.pos_labels.index(targ_dim_name)
                # Replace indices with the x, y, z values from the pixels
                pos_val_mat[:, dim_ind] = self.pos_vals_list[:, col_ind]

        ds_pos_val = MicroDataset('Position_Values', pos_val_mat)
        ds_pos_val.attrs['labels'] = pos_slice_dict
        ds_pos_val.attrs['units'] = self.pos_units

        meas_grp = MicroDataGroup(meas_grp.name, '/')
        meas_grp.addChildren([ds_pos_ind, ds_pos_val])
        
        h5_refs += self.hdf.writeData(meas_grp)
        
        # Do all the reference linking:
        aux_ds_names = ['Excitation_Waveform', 'Position_Indices', 'Position_Values', 'UDVS_Indices',
                        'Spectroscopic_Indices', 'Bin_Step', 'Bin_Indices', 'Bin_Wfm_Type',
                        'Bin_Frequencies', 'Bin_FFT', 'UDVS', 'UDVS_Labels', 'Noise_Floor', 'Spectroscopic_Values']
        linkRefs(self.ds_main, getH5DsetRefs(aux_ds_names, h5_refs))

        # While we have all the references and mean data, write the plot groups as well:
        generatePlotGroups(self.ds_main, self.hdf, self.mean_resp, 
                           self.folder_path, self.basename,
                           self.max_resp, self.min_resp, 
                           max_mem_mb=self.max_ram,
                           spec_label=self.spec_label,
                           show_plots=show_plots, save_plots=save_plots,
                           do_histogram=do_histogram)
        
        # Now that everything about this dataset is complete:
        self.dset_index += 1
        
    # ##################################################################################################
            
    def __initialize_meas_group(self, num_pix, current_pixels):
        """
        Creates and initializes the primary (and auxillary) datasets and datagroups
        to hold the raw data for the current set of experimental parameters.
        
        Parameters
        ----------
        num_pix : unsigned int
            Number of pixels this datagroup is expected to hold
        current_pixels : dictionary of BEPSndfPixel objects
            Extracted data for the first pixel in this group
            
        Returns
        ---------
        h5_refs : list of HDF5group and HDF5Dataset references 
            references of the written H5 datasets

        """

        tot_bins = 0
        tot_pts = 0
        # Each wavetype can have different number of bins
        for pixl in current_pixels.values():
            tot_bins += pixl.num_bins
            tot_pts += pixl.num_bins*pixl.num_steps
            
        # Need to halve the number of steps when only in / out field is acquired:
        if self.halve_udvs_steps:
            tot_pts = int(tot_pts/2)
        
        # Populate information from the columns within the pixels such as the FFT, bin freq, indices, etc. 
        bin_freqs = np.zeros(shape=tot_bins, dtype=np.float32)
        bin_inds = np.zeros(shape=tot_bins, dtype=np.uint32)
        bin_FFT = np.zeros(shape=tot_bins, dtype=np.complex64)
        exec_bin_vec = np.zeros(shape=tot_bins, dtype=np.int32)
        pixel_bins = {}  # Might be useful later
        stind = 0
        for wave_type in self.__unique_waves__:
            pixl = current_pixels[wave_type]
            exec_bin_vec[stind:stind+pixl.num_bins] = wave_type*np.ones(pixl.num_bins)
            bin_inds[stind:stind+pixl.num_bins] = pixl.BE_bin_ind
            bin_freqs[stind:stind+pixl.num_bins] = pixl.BE_bin_w
            bin_FFT[stind:stind+pixl.num_bins] = pixl.FFT_BE_wave
            pixel_bins[wave_type] = [stind, pixl.num_bins]
            stind += pixl.num_bins
        del pixl, stind 
                
        # Make the index matrix that has the UDVS step number and bin indices
        spec_inds = np.zeros(shape=(2, tot_pts), dtype=np.uint32)
        stind = 0
        # Need to go through the UDVS file and reconstruct chronologically
        for step_index, wave_type in enumerate(self.excit_type_vec):
            if self.halve_udvs_steps and self.udvs_mat[step_index, 2] < 1E-3:  # invalid AC amplitude
                    continue  # skip
            vals = pixel_bins[wave_type]
            spec_inds[1, stind:stind+vals[1]] = step_index * np.ones(vals[1])  # UDVS step
            spec_inds[0, stind:stind+vals[1]] = np.arange(vals[0], vals[0]+vals[1])  # Bin step
            stind += vals[1]
        del stind, wave_type, step_index
        
        self.spec_inds = spec_inds  # will need this for plot group generation

        ds_ex_wfm = MicroDataset('Excitation_Waveform',
                                 np.float32(np.real(np.fft.ifft(np.fft.ifftshift(self.BE_wave)))))
        ds_bin_freq = MicroDataset('Bin_Frequencies', bin_freqs)
        ds_bin_inds = MicroDataset('Bin_Indices', bin_inds - 1, dtype=np.uint32)  # From Matlab to Python (base 0)
        ds_bin_fft = MicroDataset('Bin_FFT', bin_FFT)
        ds_wfm_typ = MicroDataset('Bin_Wfm_Type', exec_bin_vec)
        ds_bin_steps = MicroDataset('Bin_Step', np.arange(tot_bins, dtype=np.uint32)) 
        
        curr_parm_dict = self.parm_dict
        # Some very basic information that can help the processing crew
        curr_parm_dict['num_bins'] = tot_bins
        curr_parm_dict['num_pix'] = num_pix
                
        # technically should change the date, etc.
        self.current_group = '{:s}'.format('Measurement_')
        meas_grp = MicroDataGroup(self.current_group, '/')
        meas_grp.attrs = curr_parm_dict

        chan_grp = MicroDataGroup('Channel_')
        chan_grp.attrs['Channel_Input'] = curr_parm_dict['IO_Analog_Input_1']
        meas_grp.addChildren([chan_grp])
        
        udvs_slices = dict()
        for col_ind, col_name in enumerate(self.udvs_labs):
            udvs_slices[col_name] = (slice(None), slice(col_ind, col_ind+1))
            # print('UDVS column index {} = {}'.format(col_ind,col_name))
        ds_udvs_mat = MicroDataset('UDVS', self.udvs_mat)
        ds_udvs_mat.attrs['labels'] = udvs_slices
        ds_udvs_mat.attrs['units'] = self.udvs_units
        
        actual_udvs_steps = self.num_udvs_steps
        if self.halve_udvs_steps:
            actual_udvs_steps /= 2
        if actual_udvs_steps % 1:
            raise ValueError('Actual number of UDVS steps should be an integer')
        actual_udvs_steps = int(actual_udvs_steps)
        
        curr_parm_dict['num_udvs_steps'] = actual_udvs_steps        
        
        ds_udvs_inds = MicroDataset('UDVS_Indices', self.spec_inds[1])
        # ds_udvs_inds.attrs['labels'] = {'UDVS_step':(slice(None),)}

        '''
        Create the Spectroscopic Values tables
        '''
        spec_vals, spec_inds, spec_vals_labs, spec_vals_units, spec_vals_labs_names = \
            createSpecVals(self.udvs_mat, spec_inds, bin_freqs, exec_bin_vec,
                           curr_parm_dict, np.array(self.udvs_labs), self.udvs_units)

        spec_vals_slices = dict()
        for row_ind, row_name in enumerate(spec_vals_labs):
            spec_vals_slices[row_name] = (slice(row_ind, row_ind+1), slice(None))
        ds_spec_vals_mat = MicroDataset('Spectroscopic_Values', np.array(spec_vals, dtype=np.float32))
        ds_spec_vals_mat.attrs['labels'] = spec_vals_slices
        ds_spec_vals_mat.attrs['units'] = spec_vals_units
        ds_spec_mat = MicroDataset('Spectroscopic_Indices', spec_inds, dtype=np.uint32)
        ds_spec_mat.attrs['labels'] = spec_vals_slices
        ds_spec_mat.attrs['units'] = spec_vals_units  
        for entry in spec_vals_labs_names:
            label = entry[0]+'_parameters'
            names = entry[1]
            ds_spec_mat.attrs[label] = names
            ds_spec_vals_mat.attrs[label] = names

        '''
        New Method for chunking the Main_Data dataset.  Chunking is now done in N-by-N squares of UDVS steps by pixels.
        N is determined dinamically based on the dimensions of the dataset.  Currently it is set such that individual
        chunks are less than 10kB in size.
        
        Chris Smith -- csmith55@utk.edu
        '''
        max_bins_per_pixel = np.max(pixel_bins.values())

        beps_chunks = calc_chunks([num_pix, tot_pts],
                                  np.complex64(0).itemsize,
                                  unit_chunks=(1, max_bins_per_pixel))
        ds_main_data = MicroDataset('Raw_Data',
                                    np.zeros(shape=(1, tot_pts), dtype=np.complex64),
                                    chunking=beps_chunks,
                                    resizable=True,
                                    compression='gzip')

        ds_noise = MicroDataset('Noise_Floor', np.zeros(shape=(1, actual_udvs_steps), dtype=nf32),
                                chunking=(1, actual_udvs_steps), resizable=True, compression='gzip')

        # Allocate space for the first pixel for now and write along with the complete tree...
        # Positions CANNOT be written at this time since we don't know if the parameter changed
        
        chan_grp.addChildren([ds_main_data, ds_noise, ds_ex_wfm, ds_spec_mat, ds_wfm_typ,
                              ds_bin_steps, ds_bin_inds, ds_bin_freq, ds_bin_fft, ds_udvs_mat,
                              ds_spec_vals_mat, ds_udvs_inds])
                              
        # meas_grp.showTree()
        h5_refs = self.hdf.writeData(meas_grp)
        
        self.ds_noise = getH5DsetRefs(['Noise_Floor'], h5_refs)[0] 
        self.ds_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
        self.pos_vals_list = list()
                
        # self.dset_index += 1 #  raise dset index after closing only
        self.ds_pixel_index = 0
        
        # Use this for plot groups:
        self.mean_resp = np.zeros(shape=tot_pts, dtype=np.complex64)
        
        # Used for Histograms
        self.max_resp = np.zeros(shape=num_pix, dtype=np.float32)
        self.min_resp = np.zeros(shape=num_pix, dtype=np.float32)
        
        return h5_refs
        
    # ##################################################################################################
        
    def __append_pixel_data(self, pixel_data):
        """
        Goes through the list of pixel objects and populates the raw dataset 
        and noise dataset for this spatial pixel.
        
        Parameters
        ----------
        pixel_data : List of BEPSndfPixel objects 
            List containing parsed data for this particular spatial pixel
        
        Returns
        ---------
        None

        """
        
        if self.__num_wave_types__ == 1 and not self.halve_udvs_steps:
            """Technically, this will be taken care of in the later (general) part but 
            since this condition is more common it is worth writing for specifically"""

            zero_pix = self.__unique_waves__[0]

            data_vec = pixel_data[zero_pix].spectrogram_vec
            noise_mat = np.float32(pixel_data[zero_pix].noise_floor_mat)

            # Storing a list of lists since we don't know how many pixels we will find in this measurement group
            self.pos_vals_list.append([pixel_data[zero_pix].x_value, pixel_data[zero_pix].y_value,
                                       pixel_data[zero_pix].z_value])
            
        else:

            data_vec = np.zeros(shape=(self.ds_main.shape[1]), dtype=np.complex64)
            noise_mat = np.zeros(shape=(3, self.ds_noise.shape[1]), dtype=np.float32)
            
            internal_step_index = {}
            for wave_type in self.__unique_waves__:
                internal_step_index[wave_type] = 0
                
            stind = 0
            step_counter = 0    
            # Go through each line in the UDVS file and reconstruct chronologically
            for step_index, wave_type in enumerate(self.excit_type_vec):
                # get the noise and data from correct pixel -> address by wave_number.
                
                if self.halve_udvs_steps and self.udvs_mat[step_index, 2] < 1E-3:  # invalid AC amplitude
                    # print('Step index {} was skipped'.format(step_index))
                    # Not sure why each wave type has its own counter but there must have been a good reason
                    internal_step_index[wave_type] += 1 
                    continue  # skip
                
                data_pix = pixel_data[wave_type].spectrogram_mat
                noise_pix = pixel_data[wave_type].noise_floor_mat
                enind = stind + pixel_data[wave_type].num_bins
                                
                data_vec[stind:enind] = data_pix[:, internal_step_index[wave_type]]
                noise_mat[:, step_counter] = np.float32(noise_pix[:, internal_step_index[wave_type]])
                
                stind = enind
                internal_step_index[wave_type] += 1
                step_counter += 1

            # Storing a list of lists since we don't know how many pixels we will find in this measurement group
            self.pos_vals_list.append([pixel_data[wave_type].x_value, pixel_data[wave_type].y_value,
                                       pixel_data[wave_type].z_value])
            
            del internal_step_index, stind, enind, step_index, wave_type, step_counter
        
        if self.ds_pixel_index > 0:
            # in the case of the first pixel, we already reserved zeros- no extension
            # for all other lines - we extend the dataset before writing
            self.ds_main.resize(self.ds_main.shape[0]+1, axis=0)
            self.ds_noise.resize(self.ds_noise.shape[0]+1, axis=0)

        self.ds_main[-1, :] = data_vec
        self.ds_noise[-1] = np.array([tuple(noise) for noise in noise_mat.T], dtype=nf32)

        self.hdf.file.flush()
        
        # Take mean response here:
        self.mean_resp = (1/(self.ds_pixel_index + 1))*(data_vec + self.ds_pixel_index*self.mean_resp)
        
        self.max_resp[self.ds_pixel_index] = np.amax(np.abs(data_vec))
        self.min_resp[self.ds_pixel_index] = np.amin(np.abs(data_vec))
        
        self.ds_pixel_index += 1
               
    ###################################################################################################
    
    def _parse_file_path(self, file_path):
        """
        Returns the file paths to the parms text file and UDVS spreadsheet.\n
        Note: This function also initializes the basename and the folder_path for this instance
        
        Parameters
        ----------
        file_path : String / unicode
            Absolute path of the any file inside the measurment folder
        
        Returns
        -------
        parm_filepath : String / unicode
            absolute filepath of the parms text file
        udvs_filepath : String / unicode
            absolute file path of the UDVS spreadsheet
        parms_mat_path : String / unicode
            absolute filepath of the .mat parms file

        """
        udvs_filepath = None
        folder_path, tail = path.split(file_path)
        main_folder_path = None
        for item in listdir(folder_path):
            if item == 'newdataformat':
                main_folder_path = folder_path
                self.folder_path = path.join(folder_path, item)
                break
        if main_folder_path is None:
            self.folder_path = folder_path
            # need to set the parent as the root_folder
            main_folder_path, tail = path.split(folder_path)

        parms_mat_path = None
        parm_filepath = None
        for filenames in listdir(main_folder_path):
            if filenames.endswith('.txt') and filenames.find('parm') > 0:
                parm_filepath = path.join(main_folder_path, filenames)
            elif filenames.endswith('all.mat'):
                parms_mat_path = path.join(main_folder_path, filenames)
            elif filenames.endswith('more_parms.mat') and parms_mat_path is None:
                parms_mat_path = path.join(main_folder_path, filenames)
        for filenames in listdir(self.folder_path):
            if (filenames.endswith('.xlsx') or filenames.endswith('.xls')) and filenames.find('UD_VS') > 0:
                udvs_filepath = path.join(self.folder_path, filenames)
                break
        if parm_filepath is None:
            for filenames in listdir(self.folder_path):
                if (filenames.endswith('.xls') or filenames.endswith('.xlsx')) and filenames.find('parm') > 0:
                    parm_filepath = path.join(main_folder_path, filenames)
                
        (tail, self.basename) = path.split(main_folder_path)
        
        return parm_filepath, udvs_filepath, parms_mat_path
        
    ###################################################################################################

    def __assemble_parsers(self):
        """
        Returns a list of BEPSndfParser objects per excitation wave type
        
        Returns
        ---------
        parsers: list of BEPSndfParser objects 
            list of the same length as the input numpy array
        """
        parsers = []  # Maybe this needs to be a dictionary instead for easier access?
        for wave_type in self.__unique_waves__:
            filename = self.basename + '_1_'
            if wave_type > 0:
                filename = self.basename + '_1_' + str(wave_type) + '.dat'
            else:
                filename = self.basename + '_1_r' + str(abs(wave_type)) + '.dat'
            datapath = path.join(self.folder_path, filename)
            if not path.isfile(datapath):
                raise LookupError('Error!!: {}expected but not found!'.format(filename))
                # return
            parsers.append(BEPSndfParser(datapath, wave_type))
        return parsers
        
    # ##################################################################################################
        
    @staticmethod
    def __get_excit_wfm(filepath):
        """
        Returns the excitation BE waveform present in the more parms.mat file
        
        Parameters
        ------------
        filepath : String / unicode
            Absolute filepath of the .mat parameter file
        
        Returns
        -----------
        ex_wfm : 1D numpy float array
            Band Excitation waveform

        """
        if not path.exists(filepath):
            warn('BEPSndfTranslator - NO more_parms.mat file found')
            return np.zeros(1000, dtype=np.float32)

        if 'more_parms' in filepath:
            matread = loadmat(filepath, variable_names=['FFT_BE_wave'])
            fft_full = np.complex64(np.squeeze(matread['FFT_BE_wave']))
            bin_inds = None
            fft_full_rev = None
        else:
            matread = loadmat(filepath, variable_names=['FFT_BE_wave', 'FFT_BE_rev_wave', 'BE_bin_ind'])
            bin_inds = np.uint(np.squeeze(matread['BE_bin_ind']))-1
            fft_full = np.complex64(np.squeeze(matread['FFT_BE_wave']))
            fft_full_rev = np.complex64(np.squeeze(matread['FFT_BE_rev_wave']))

        return fft_full, fft_full_rev, bin_inds

    # ##################################################################################################
        
    @staticmethod
    def __read_udvs_table(udvs_filepath):
        """
        Reads the UDVS spreadsheet in either .xls or .xlsx format!
        
        Parameters
        ----------
        udvs_filepath : String / unicode
            absolute path to the spreadsheet
            
        Returns
        ----------
        udvs_labs : list of strings
            names of columns in the UDVS table
        udvs_units : list of strings
            units for columns in the UDVS table
        UDVS_mat : 2D numpy float array
            Contents of the UDVS table

        """
        workbook = xlreader.open_workbook(udvs_filepath)
        worksheet = workbook.sheet_by_index(0)
        udvs_labs = list()
        for col in range(worksheet.ncols):
            udvs_labs.append(str(worksheet.cell(0, col).value))
        # sometimes, the first few columns are named incorrectly. FORCE them to be named correclty:
        udvs_units = list(['' for _ in range(len(udvs_labs))])
        udvs_labs[0:5] = ['step_num', 'dc_offset', 'ac_amp', 'wave_type', 'wave_mod']
        udvs_units[0:5] = ['', 'V', 'A', '', '']
        
        udvs_mat = np.zeros(shape=(worksheet.nrows-1, worksheet.ncols), dtype=np.float32)
        for row in range(1, worksheet.nrows):
            for col in range(worksheet.ncols):
                try:
                    udvs_mat[row-1, col] = worksheet.cell(row, col).value
                except ValueError:
                    udvs_mat[row-1, col] = float('NaN')
                except:
                    raise
                    
        # Decrease counter of number of steps by 1 (Python base 0)
        udvs_mat[:, 0] -= 1
        
        return udvs_labs, udvs_units, udvs_mat

    @staticmethod
    def __get_unique_wave_types(vec):
        """
        Returns a numpy array containing the different waveforms in a 
        BEPS experiment in the format: [1,-1,2,-2.....]
        
        Parameters
        ---------------
        vec : 1D list or 1D numpy array 
            waveform types in the UDVS table
            
        Returns
        ----------
        uniq : 1D numpy array
            Unique waveform types in format listed above

        """
        sorted_all = np.unique(vec)
        pos_vals = sorted_all[sorted_all >= 0]
        neg_vals = sorted_all[sorted_all < 0]
        
        if len(pos_vals) == 0:
            return neg_vals
            
        if len(neg_vals) == 0:
            return pos_vals
            
        uniq = []
        posind = 0 
        negind = len(neg_vals)-1
        while posind < len(pos_vals) or negind >= 0:
            
            if posind == len(pos_vals):
                uniq += list(neg_vals)
                break
            if negind == len(neg_vals):
                uniq += list(pos_vals)
                break
            # Otherwise compare
            if pos_vals[posind] < abs(neg_vals[negind]):
                uniq.append(pos_vals[posind])
                posind += 1
            elif pos_vals[posind] == abs(neg_vals[negind]):
                uniq.append(pos_vals[posind])
                uniq.append(neg_vals[negind])
                posind += 1
                negind -= 1
            else:
                uniq.append(neg_vals[negind])
                negind -= 1
                        
        return np.array(uniq)


class BEPSndfParser(object):
    """
    An object of this class is given the responsibility to step through a 
    BEPS new data format file and return parsed BEPSndfPixel objects.\n
    This class is NOT responsible for actually parsing the byte contents within
    each pixel.\n
    Each wave type is given its own Parser object since it has a file of its own
    """
    
    def __init__(self, file_path, wave_type=1, scout=True):
        """
        Initializes the BEPSndfParser object with following inputs:
        
        Parameters
        -----------
        file_path : string or unicode
            Absolute path of the .dat file
        wave_type : int (optional. Default = 1)
            Integer value signifying type of the excitation waveform\n
        scout : Boolean (optional. Default = true) 
            whether or not the parser should figure out basic details such as 
            the number of pixels, and the spatial dimensionality

        """
        self.__file_handle__ = open(file_path, "rb")
        self.__EOF__ = False
        self.__curr_Pixel__ = 0
        self.__start_point__ = 0
        self.__wave_type__ = wave_type
        self.__filesize__ = path.getsize(file_path)
        self.__pixel_indices__ = list()
        if scout:
            self.__scout()
        
    def get_wave_type(self):
        """
        Returns the excitation wave type as an integer
        
        Returns
        -------
        wave_type : int
            Wave type. Positive number means chirp up, negative number is chirp down.

        """
        return self.__wave_type__
        
    def get_num_pixels(self):
        """
        Returns the total number of spatial pixels. This includes X, Y, Z, Laser positions
        
        Returns
        -------
        num_pix : unsigned int
            Number of pixels in this file

        """
        return self.__num_pixels__
        
    def get_spatial_pixels(self):
        """
        Returns the number of steps in each spatial dimension 
        organized from fastest to slowest varying dimension
        
        Returns
        -------
        __num_laser_steps__ : unsigned int
            Number of laser steps
        __num_z_steps__ : unsigned int
            Number of height steps
        __num_x_steps__ : unsigned int
            Number of columns
        __num_y_steps__ : unsigned int
            Number of rows
        """
        return self.__num_laser_steps__, self.__num_z_steps__, self.__num_x_steps__, self.__num_y_steps__
    
    # Don't use this to figure out if something changes. You need pixel to previous pixel comparison    
    def __scout(self):
        """
        Steps through the file quickly without parsing it. 
        The idea is to calculate the number of pixels ahead of time so that 
        it is easier to parse the dataset. 
        For phase checking, it is recommended that this function be modified to 
        also keep track of the byte positions of the pixels so that pixels can be 
        directly accessed if need be.

        """
        count = 0
        self.__num_pixels__ = 0
        while True:
            self.__pixel_indices__.append(self.__start_point__*4)
            self.__file_handle__.seek(self.__start_point__*4, 0)
            spectrogram_length = int(np.fromstring(self.__file_handle__.read(4), dtype='f')[0])  # length of spectrogram
            
            if count == 0:
                self.__file_handle__.seek(self.__start_point__*4, 0)
                data_vec = np.fromstring(self.__file_handle__.read(spectrogram_length*4), dtype='f')
                pix = BEPSndfPixel(data_vec, self.__wave_type__)
                self.__num_x_steps__ = pix.num_x_steps
                self.__num_y_steps__ = pix.num_y_steps
                self.__num_z_steps__ = pix.num_z_steps
                self.__num_bins__ = pix.num_bins
                
            count += 1
            self.__start_point__ += spectrogram_length
            
            if self.__filesize__ == self.__start_point__*4:
                self.__num_pixels__ = count
                
                # Laser position spectroscopy is NOT accounted for anywhere. 
                # It is impossible to find out from the parms.txt, UD_VS, or the binary .dat file
                num_laser_steps = 1.0*count/(self.__num_z_steps__*self.__num_y_steps__*self.__num_x_steps__)                
                if num_laser_steps % 1.0 != 0:
                    print('Some parameter changed inbetween. \
                          BEPS NDF Translator does not handle this usecase at the moment')
                else:
                    self.__num_laser_steps__ = int(num_laser_steps)
                
                break
            
            if self.__filesize__ > self.__start_point__*4:
                self.__num_pixels__ = -1
                
        self.__start_point__ = 0
        
        spat_dim = 0
        if self.__num_z_steps__ > 1:
            # print('Z is varying')
            spat_dim += 1
        if self.__num_y_steps__ > 1:
            # print('Y is varying')
            spat_dim += 1
        if self.__num_x_steps__ > 1:
            # print('X is varying')
            spat_dim += 1
        if self.__num_laser_steps__ > 1:
            # Laser spot position vector is junk in the .dat file
            # print('Laser position / unknown parameter varying')
            spat_dim += 1
        # print('Total of {} spatial dimensions'.format(spat_dim))
        self.__spat_dim__ = spat_dim
             
    def read_pixel(self, bin_fft=None):
        """
        Returns a BEpixel object containing the parsed information within a pixel.
        Moves pixel index up by one.
        This is where one could conceivably read the file in one pass instead of making 100,000 file I/Os.

        Returns
        -------
        pixel : BEPSndfPixel
            Object that describes the data contained within the pixel
        """
        
        if self.__filesize__ == self.__start_point__*4:
            print('BEPS NDF Parser - No more pixels left!')
            return -1
        
        self.__file_handle__.seek(self.__start_point__ * 4, 0)
        spectrogram_length = int(np.fromstring(self.__file_handle__.read(4), dtype='f')[0])  # length of spectrogram
        self.__file_handle__.seek(self.__start_point__ * 4, 0)
        data_vec = np.fromstring(self.__file_handle__.read(spectrogram_length*4), dtype='f')
       
        self.__start_point__ += spectrogram_length
        self.__curr_Pixel__ += 1
                
        if self.__filesize__ == self.__start_point__*4:
            print('BEPS NDF Parser reached End of File')
            self.__EOF__ = True
            self.__file_handle__.close()
                
        return BEPSndfPixel(data_vec, abs(self.__wave_type__), bin_fft)
        

class BEPSndfPixel(object):
    """
    Stands for BEPS (new data format) Pixel. 
    This class parses (and keeps) the stream of data contained in a single cell of a BEPS data set of the new data 
    format. Access desired parameter directly without get methods.
    """
    
    def __init__(self, data_vec, harm=1, bin_fft=None):
        """
        Initializes the pixel instance by parsing the provided data. 
        
        Parameters
        ----------
        data_vec : 1D float numpy array
            Data contained within each pixel
        harm: unsigned int
            Harmonic of the BE waveform. absolute value of the wave type used to normalize the response waveform.

        """
        
        harm = abs(harm)
        if harm > 3 or harm < 1:
            harm = 1
            warn('Error in BEPSndfPixel: invalid wave type / harmonic provided.')
        self.harm = harm
        # Begin parsing data:
        self.spatial_index = int(data_vec[1])-1
        
        self.spectrogram_length = int(data_vec[0]) 
        
        # calculate indices for parsing
        s1 = int(data_vec[2])  # total rows in pixel
        s2 = int(data_vec[3])  # total cols in pixel
        data_vec1 = data_vec[2:self.spectrogram_length]    
        data_mat1 = data_vec1.reshape(s1, s2)
        spect_size1 = int(data_mat1[1, 0])  # total rows in spectrogram set
        self.num_bins = int(spect_size1/2)   # or, len(BE_bin_w)
        self.num_steps = int(data_mat1[1, 1])  # total cols in spectrogram set 
        s3 = int(s1-spect_size1)  # row index of beginning of spectrogram set
        s4 = int(s2-self.num_steps)  # col index of beginning of spectrogram set
            
        self.wave_label = data_mat1[2, 0]  # This is useless
        self.wave_modulation_type = data_mat1[2, 1]  # this is the one with useful information
        # print 'Pixel #',self.spatial_index,' Wave label: ',self.wave_label, ', Wave Type: ', self.wave_modulation_type
        
        # First get the information from the columns:   
        fft_be_wave_real = data_mat1[s3:s3-0+self.num_bins, 1]  # real part of excitation waveform  
        fft_be_wave_imag = data_mat1[s3+self.num_bins:s3-0+spect_size1, 1]  # imaginary part of excitation waveform  
        
        """ Though typecasting the combination of the real and imaginary data looks fine in HDFviewer and Spyder, 
        Labview sees such data as an array of clusters having 'r' and 'i' elements """
        # self.FFT_BE_wave = np.complex64(fft_be_wave_real + 1j*fft_be_wave_imag) 
        
        # complex excitation waveform! due to a problem in the acquisition software, this may not be normalized properly
        self.FFT_BE_wave = np.zeros(self.num_bins, dtype=np.complex64)
        self.FFT_BE_wave.real = fft_be_wave_real
        self.FFT_BE_wave.imag = fft_be_wave_imag
        
        del fft_be_wave_real, fft_be_wave_imag            
        
        self.BE_bin_w = data_mat1[s3:s3-0+self.num_bins, 2]  # vector of band frequencies
        # vector of band indices (out of all accessible frequencies below Nyquist frequency)
        self.BE_bin_ind = data_mat1[s3+self.num_bins:s3-0+spect_size1, 2] 
     
        # Now look at the top few rows to get more information: 
        self.daq_channel = data_mat1[2, 2]
        # self.num_x_steps = int(data_mat1[3, 0])
        # self.num_y_steps = int(data_mat1[4, 0])
        self.num_z_steps = int(data_mat1[5, 0])
        self.z_index = int(data_mat1[5, 1] - 1)
        # self.y_index = int(data_mat1[4, 1] - 1)
        # self.x_index = int(data_mat1[3, 1] - 1)
        self.z_value = data_mat1[5, 2]
        # self.y_value = data_mat1[4, 2]
        # self.x_value = data_mat1[3, 2]
        self.num_x_steps = int(data_mat1[4, 0])
        self.num_y_steps = int(data_mat1[3, 0])
        self.x_index = int(data_mat1[4, 1] - 1)
        self.y_index = int(data_mat1[3, 1] - 1)
        self.x_value = data_mat1[4, 2]
        self.y_value = data_mat1[3, 2]

        self.step_ind_vec = data_mat1[0, s4:]  # vector of step indices   
        self.DC_off_vec = data_mat1[1, s4:]  # vector of dc offsets  voltages
        self.AC_amp_vec = data_mat1[2, s4:]  # vector of ac amplitude voltages 

        # matrix of noise floor data. Use this information to exclude bins during fitting
        self.noise_floor_mat = data_mat1[3:6, s4:]
        
        # self.plot_group_list_mat = data_mat1[6:s3-2, s4:]  # matrix of plot groups
        
        # Here come the optional parameter rows:
        self.deflVolt_vec = data_mat1[s3-2, s4:]  # vector of dc cantilever deflection
        # I think this is how the defl setpoint vec should be fixed:
        self.deflVolt_vec[np.isnan(self.deflVolt_vec)] = 0
        # so far, this vector seemed to match the DC offset vector....
        
        self.laser_spot_pos_vec = data_mat1[s3-1, s4:]  # NEVER used
        
        # Actual data for this pixel:
        spectrogram_real_mat = data_mat1[s3:s3+self.num_bins, s4:]  # real part of response spectrogram
        spectrogram_imag_mat = data_mat1[s3+self.num_bins:s3+spect_size1, s4:]  # imaginary part of response spectrogram
        # Be consistent and ensure that the data is also stored as 64 bit complex as in the array creation
        # complex part of response spectrogram
        self.spectrogram_mat = np.complex64(spectrogram_real_mat + 1j*spectrogram_imag_mat)  
        del spectrogram_real_mat, spectrogram_imag_mat 

        if bin_fft is not None:
            self.FFT_BE_wave = bin_fft

        self.spectrogram_mat = normalizeBEresponse(self.spectrogram_mat, self.FFT_BE_wave, harm)

        #  Reshape as one column (its free in Python anyway):
        temp_mat = self.spectrogram_mat.transpose() 
        self.spectrogram_vec = temp_mat.reshape(self.spectrogram_mat.size)

    def is_different_from(self, prev_pixel):
        """
        Compares parameters in this object with those another BEPSndfPixel object 
        to tell if any parameter has changed between these pixels
        
        Parameters
        ----------
        prev_pixel : BEPSndfPixel object
            The other pixel object to compare this pixel to
            
        Returns
        -------
        is_different : Boolean
            Whether or not these pixel objects share the same parameters

        Notes
        -----
        *Typical things that change during BEPS*
        1. BE parameters:
        a. Center Frequency, Band Width - changes in the BE_bin_w
        b. Amplitude, Phase Variation, Band Edge Smoothing, Band Edge Trim - Harder to find out what happened
        exactly - FFT should show changes
        c. BE repeats, desired duration - changes in the spectrogram length?
        2. VS Parameters
        a. Amplitude, Phase shift - Changes in the AC_amp_vec / DC offset
        b. Offset, Read voltage - Shows up in the DC offset
        c. Steps per full Cycle - Changes in DC offset / AC amplitude ....
        d. Number of cycles, Cycle fraction, In and out of field - changes in the length of DC offset etc.
        e. FORC - should all show up in DC / AC amplitude
        3. IO parameters : don't change really
        4. grid parameters : cannot do anything about this.

        """
        disp_on = True        
        
        if self.spectrogram_length != prev_pixel.spectrogram_length:
            if disp_on:
                print('Spectrogram Length changed on pixel {}'.format(self.spatial_index))
            return True
                        
        if self.num_bins != prev_pixel.num_bins:
            if disp_on:
                print('Number of bins changed on on pixel {}'.format(self.spatial_index))
            return True
            
        if not np.array_equal(self.BE_bin_w, prev_pixel.BE_bin_w):
            if disp_on:
                print('Bin Frequencies changed on pixel {}'.format(self.spatial_index))
            return True
            
        if not np.array_equal(self.FFT_BE_wave, prev_pixel.FFT_BE_wave):
            if disp_on:
                print('BE FFT changed on pixel {}'.format(self.spatial_index))
            return True
            
        if not np.array_equal(self.AC_amp_vec, prev_pixel.AC_amp_vec):
            if disp_on:
                print('AC amplitude (UDVS) changed on pixel {}'.format(self.spatial_index))
            return True
        
        if not np.array_equal(self.DC_off_vec, prev_pixel.DC_off_vec):
            if disp_on:
                print('DC offset (UDVS) changed on pixel {}'.format(self.spatial_index))
            return True
        
        # I was told that this section was just garbage in the file.
        # if not np.array_equal(self.laser_spot_pos_vec, prev_pixel.laser_spot_pos_vec):
            # print 'Laser spot position vec was different....'
            # print self.laser_spot_pos_vec
            # return True
            
        if not np.array_equal(self.deflVolt_vec, prev_pixel.deflVolt_vec):
            print('deflVolt_vec vec was different....')
            return True
        
        return False



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 21 12:29:33 2015
Main Class in charge of writing/reading to/from hdf5 file.
@author: Numan Laanait, Suhas Somnath, Chris Smith
"""

# cannot import unicode_literals since it is not compatible with h5py just yet
from __future__ import division, print_function, absolute_import, unicode_literals
import os
import subprocess
import sys
from collections import Iterable
from time import time, sleep
from warnings import warn

import h5py
import numpy as np

from .microdata import MicroDataGroup
from ..__version__ import version

if sys.version_info.major == 3:
    unicode = str


class ioHDF5(object):

    def __init__(self, file_handle, cachemult=1):
        """
        Handles:
            + I/O operation from HDF5 file.
            + Utilities to get data and associated auxiliary.

        Parameters
        ----------
        file_handle : Object - String or Unicode or open hdf5 file
            Absolute path to the h5 file or an open hdf5 file
        cachemult : unsigned int (Optional. default = 1)
            Cache multiplier
        """
        if type(file_handle) in [str, unicode]:
            # file handle is actually a file path
            # propfaid = h5py.h5p.create(h5py.h5p.FILE_ACCESS)
            # if cachemult != 1:
            #     settings = list(propfaid.get_cache())
            #     settings[2] *= cachemult
            #     propfaid.set_cache(*settings)
            # try:
            #     fid = h5py.h5f.open(file_handle, fapl=propfaid)
            #     self.file = h5py.File(fid, mode = 'r+')
            # except IOError:
            #     #print('Unable to open file %s. \n Making a new one! \n' %(filename))
            #     fid = h5py.h5f.create(file_handle, fapl=propfaid)
            #     self.file = h5py.File(fid, mode = 'w')
            # except:
            #     raise
            try:
                self.file = h5py.File(file_handle, 'r+')
            except IOError:
                self.file = h5py.File(file_handle, 'w')
            except:
                raise
            
            self.path = file_handle
        elif type(file_handle) == h5py.File:
            # file handle is actually an open hdf file
            if file_handle.mode == 'r':
                warn('ioHDF5 cannot work with open HDF5 files in read mode. Change to r+ or w')
                return
            self.file = file_handle.file
            self.path = file_handle.filename

    def clear(self):
        """
        Clear h5.file of all contents

        file.clear() only removes the contents, it does not free up previously allocated space.
        To do so, it's necessary to use the h5repack command after clearing.
        Because the file must be closed and reopened, it is best to call this
        function immediately after the creation of the ioHDF5 object.
        """
        self.file.clear()
        self.repack()

    def repack(self):
        """
        Uses the h5repack command to recover cleared space in an hdf5 file.
        h5repack can also be used to change chunking and compression, but these options have
        not yet been implemented here.
        """
        self.close()
        tmpfile = self.path+'.tmp'

        '''
        Repack the opened hdf5 file into a temporary file
        '''
        try:
            repack_line = ' '.join(['h5repack', '"'+self.path+'"', '"'+tmpfile+'"'])
            subprocess.check_output(repack_line,
                                    stderr=subprocess.STDOUT,
                                    shell=True)
            # Check that the file is done being modified
            sleep(0.5)
            while time()-os.stat(tmpfile).st_mtime <= 1:
                sleep(0.5)
        except subprocess.CalledProcessError as err:
            print('Could not repack hdf5 file')
            raise Exception(err.output)
        except:
            raise

        '''
        Delete the original file and move the temporary file to the originals path
        '''
        # TODO Find way to get the real OS error that works in and out of Spyder
        try:
            os.remove(self.path)
            os.rename(tmpfile, self.path)
        except:
            print('Could not copy repacked file to original path.')
            print('The original file is located {}'.format(self.path))
            print('The repacked file is located {}'.format(tmpfile))
            raise

        '''
        Open the repacked file
        '''
        self.file = h5py.File(self.path, mode='r+')

    def close(self):
        """
        Close h5.file
        """
        self.file.close()

    def delete(self):
        """
        Delete h5.file
        """
        self.close()
        os.remove(self.path)
        self.file = h5py.File(self.path, 'w')

    def flush(self):
        """
        Flush data from memory and commit to file.
        Use this after manually inserting data into the hdf dataset
        """
        self.file.flush()

    def writeData(self, data, print_log=False):
        """
        Writes data into the hdf5 file and assigns data attributes such as region references.
        The tree structure is inferred from the AFMData Object.

        Parameters
        ----------
        data : Instance of MicroData
            Tree structure describing the organization of the data
        print_log : Boolean (Optional)
            Whether or not to print all log statements - use for debugging purposes

        Returns
        -------
        refList : List of HDF5dataset or HDF5Datagroup references
            References to the objects written
        """

        h5_file = self.file

        h5_file.attrs['Pycroscopy version'] = version

        # Checking if the data is an MicroDataGroup object
        if not isinstance(data, MicroDataGroup):
            warn('Input of type: {} \n'.format(type(data)))
            sys.exit("Input not of type MicroDataGroup.\n We're done here! \n")

        # Figuring out if the first item in AFMData tree is file or group
        if data.name == '' and data.parent == '/':
            # For file we just write the attributes

            for key, val in data.attrs.items():
                h5_file.attrs[key] = self.clean_string_att(val)
            if print_log:
                print('Wrote attributes of file {} \n'.format(h5_file.name))
            root = h5_file.name
        else:
            # For a group we write it and its attributes
            if data.indexed:
                ''' If the name of the requested group ends in a '_', the user expects
                the suffix index to be appended automatically. Here, we check to
                ensure that the chosen index is new.
                '''
                previous = list()
                for key in h5_file[data.parent].keys():
                    if data.name in key:
                        previous.append(key)
                if len(previous) == 0:
                    index = 0
                else:
                    # assuming that the last element of previous contains the highest index
                    last = h5_file[data.parent][previous[-1]].name
                    index = int(last.split('_')[-1])+1
                data.name += '{:03d}'.format(index)
            try:
                g = h5_file[data.parent].create_group(data.name)
                if print_log:
                    print('Created group {}'.format(g.name))
            except ValueError:
                g = h5_file[data.parent][data.name]
                if print_log:
                    print('Group already exists: {}'.format(g.name))
            except:
                h5_file.flush()
                h5_file.close()
                raise
            for key, val in data.attrs.items():
                if val is None:
                    continue
                if print_log:
                    print('Writing attribute: {} with value: {}'.format(key, val))
                g.attrs[key] = self.clean_string_att(val)
            if print_log:
                print('Wrote attributes to group: {} \n'.format(data.name))
            root = g.name

        # Populating the tree structure recursively
        ref_list = []
        # Recursive function

        def __populate(child, parent):
            """
            Recursive function to build the tree from the top down.

            Parameters
            ----------
            child
            parent

            Returns
            -------

            """
            # Update the parent attribute with the true path
            child.parent = parent

            if isinstance(child, MicroDataGroup):
                if child.indexed:
                    previous = np.where([child.name in key for key in h5_file[parent].keys()])[0]
                    if len(previous) == 0:
                        index = 0
                    else:
                        last = h5_file[parent].keys()[previous[-1]]
                        index = int(last.split('_')[-1])+1
                    child.name += '{:03d}'.format(index)
                try:
                    itm = h5_file[parent].create_group(child.name)
                    if print_log:
                        print('Created Group {}'.format(itm.name))
                except ValueError:
                    itm = h5_file[parent][child.name]
                    print('Found Group already exists {}'.format(itm.name))
                except:
                    h5_file.flush()
                    h5_file.close()
                    raise
                for key, val in child.attrs.items():
                    if val is None:
                        continue
                    if print_log:
                        print('Writing attribute: {} with value: {}'.format(key, val))
                    itm.attrs[key] = self.clean_string_att(val)
                if print_log:
                    print('Wrote attributes to group {}\n'.format(itm.name))
                # here we do the recursive function call
                for ch in child.children:
                    __populate(ch, parent+'/'+child.name)
            else:
                if not child.resizable:
                    if not bool(child.maxshape):
                        # finite sized dataset and maxshape is not provided
                        # Typically for small / ancilliary datasets
                        try:
                            itm = h5_file[parent].create_dataset(child.name,
                                                                 data=child.data,
                                                                 compression=child.compression,
                                                                 dtype=child.data.dtype,
                                                                 chunks=child.chunking)
                        except RuntimeError:
                            itm = h5_file[parent][child.name]
                            warn('Found Dataset already exists {}'.format(itm.name))
                        except:
                            h5_file.flush()
                            h5_file.close()
                            raise
                    else:
                        # In many cases, we DON'T need resizable datasets but we know the max-size
                        # Here, we only allocate the space. The provided data is ignored
                        # print child.name
                        try:
                            itm = h5_file[parent].create_dataset(child.name, child.maxshape,
                                                                 compression=child.compression,
                                                                 dtype=child.dtype,
                                                                 chunks=child.chunking)
                        except RuntimeError:
                            itm = h5_file[parent][child.name]
                            warn('Found Dataset already exists {}'.format(itm.name))
                        except:
                            h5_file.flush()
                            h5_file.close()
                            raise
                else:
                    # Resizable but the written files are significantly larger
                    max_shape = tuple([None for _ in range(len(child.data.shape))])
                    try:
                        itm = h5_file[parent].create_dataset(child.name,
                                                             data=child.data,
                                                             compression=child.compression,
                                                             dtype=child.data.dtype,
                                                             chunks=child.chunking,
                                                             maxshape=max_shape)
                    except RuntimeError:
                        itm = h5_file[parent][child.name]
                        warn('Found Dataset already exists {}'.format(itm.name))
                    except:
                        h5_file.flush()
                        h5_file.close()
                        raise

                if print_log:
                    print('Created Dataset {}'.format(itm.name))
                for key, val in child.attrs.items():
                    # print('Found some region references')
                    # writing region reference
                    if key == 'labels':
                        # print('Found some region references')
                        labels = child.attrs[key]  # labels here is a dictionary
                        self.write_region_references(itm, labels, print_log=print_log)
                        '''
                        Now make an attribute called 'labels' that is a list of strings 
                        First ascertain the dimension of the slicing:
                        '''
                        found_dim = False
                        for dimen, slobj in enumerate(list(labels.values())[0]):
                            # We make the assumption that checking the start is sufficient
                            if slobj.start is not None:
                                found_dim = True
                                break
                        if found_dim:
                            headers = [None]*len(labels)  # The list that will hold all the names
                            for col_name in labels.keys():
                                headers[labels[col_name][dimen].start] = col_name
                            if print_log:
                                print('Writing header attributes: {}'.format(key))
                            # Now write the list of col / row names as an attribute:
                            itm.attrs[key] = self.clean_string_att(headers)
                        else:
                            warn('Unable to write region labels for %s' % (itm.name.split('/')[-1]))

                        if print_log:
                            print('Wrote Region References of Dataset %s' % (itm.name.split('/')[-1]))
                    else:
                        if print_log:
                            print('Writing attribute: {} with value: {}'.format(key, val))
                        itm.attrs[key] = self.clean_string_att(child.attrs[key])
                        if print_log:
                            print('Wrote Attributes of Dataset %s \n' % (itm.name.split('/')[-1]))
                        # Make a dictionary of references
            ref_list.append(itm)
            return ref_list

        # Recursive function is called at each stage beginning at the root
        for child in data.children:
            __populate(child, root)

        if print_log:
            print('Finished writing to h5 file.\n' +
                  'Right now you got yourself a fancy folder structure. \n' +
                  'Make sure you do some reference linking to take advantage of the full power of HDF5.')
        return ref_list

    @staticmethod
    def clean_string_att(att_val):
        """
        Replaces any unicode objects within lists with their string counterparts to ensure compatibility with python 3.
        If the attribute is indeed a list of unicodes, the changes will be made in-place

        Parameters
        ----------
        att_val : object
            Attribute object

        Returns
        -------
        att_val : object
            Attribute object
        """
        try:
            if isinstance(att_val, Iterable):
                if np.any([type(x) in [str, bytes] for x in att_val]):
                    return np.array(att_val, dtype='S')
            if type(att_val) == np.str_:
                return str(att_val)
            return att_val
        except TypeError:
            warn('Failed to clean: {}'.format(att_val))
            raise

    @staticmethod
    def write_region_references(dataset, slices, print_log=False):
        """
        Creates attributes of a h5.Dataset that refer to regions in the arrays

        Parameters
        ----------
        dataset : h5.Dataset instance
            Dataset to which region references will be added as attributes
        slices : dictionary
            The slicing information must be formatted using tuples of slice objects.
            For example {'region_1':(slice(None, None), slice (0,1))}
        print_log : Boolean (Optional. Default = False)
            Whether or not to print status messages
        """
        if print_log: print('Starting to write Region References to Dataset', dataset.name, 'of shape:', dataset.shape)
        for sl in slices.keys():
            if print_log:
                print('About to write region reference:', sl, ':', slices[sl])
            if len(slices[sl]) == len(dataset.shape):
                dataset.attrs[sl] = dataset.regionref[slices[sl]]
                if print_log:
                    print('Wrote Region Reference:%s' % sl)
            else:
                warn('Region reference %s could not be written since the object size was not equal to the dimensions of'
                     ' the dataset' % sl)
                raise ValueError



from . import be_odf
from . import be_odf_relaxation
from . import beps_ndf
from . import general_dynamic_mode
from . import gmode_iv
from . import gmode_line
from . import gmode_tune
from . import image
from . import ndata_translator
from . import numpy_translator
from . import trKPFM_translator
from . import igor_ibw
from . import oneview
from . import ptychography
from . import sporc
from . import time_series
from . import translator
from . import utils
from . import df_utils
from . import beps_data_generator

from .be_odf import BEodfTranslator
from .be_odf_relaxation import BEodfRelaxationTranslator
from .beps_ndf import BEPSndfTranslator
from .general_dynamic_mode import GDMTranslator
from .gmode_iv import GIVTranslator
from .gmode_line import GLineTranslator
from .gmode_tune import GTuneTranslator
from .igor_ibw import IgorIBWTranslator
from .image import ImageTranslator
from .ndata_translator import NDataTranslator
from .numpy_translator import NumpyTranslator
from .trKPFM_translator import TRKPFMTranslator
from .oneview import OneViewTranslator
from .ptychography import PtychographyTranslator
from .sporc import SporcTranslator
from .time_series import MovieTranslator
from .translator import Translator
from .beps_data_generator import FakeDataGenerator
from .labview_h5_patcher import LabViewH5Patcher

__all__ = ['Translator', 'BEodfTranslator', 'BEPSndfTranslator', 'BEodfRelaxationTranslator',
           'GIVTranslator', 'GLineTranslator', 'GTuneTranslator', 'GDMTranslator', 'PtychographyTranslator',
           'SporcTranslator', 'MovieTranslator', 'IgorIBWTranslator', 'NumpyTranslator',
           'OneViewTranslator', 'ImageTranslator', 'NDataTranslator', 'FakeDataGenerator',
           'LabViewH5Patcher']



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 28 12:50:47 2016

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import numpy as np  # For array operations
from os import path
from .numpy_translator import NumpyTranslator


class AscTranslator(NumpyTranslator):
    """
    Translates Scanning Tunnelling Spectroscopy (STS) data in .asc files obtained from Omicron STMs.
    """

    def translate(self, file_path, max_v=1):
        """
        Translates the provided .asc file to .h5

        Parameters
        ----------
        file_path : string / unicode
            Absolute path of the source .ASC STS file from Omicron STMs
        max_v : float


        Returns
        -------
        h5_path : string / unicode
            Absolute path of the translated file
        """

        max_v = abs(max_v)

        folder_path, file_name = path.split(file_path)
        file_name = file_name[:-4]

        # Extracting the raw data into memory
        file_handle = open(file_path, 'r')
        string_lines = file_handle.readlines()
        file_handle.close()

        # Extract parameters from the first few header lines
        parm_dict = self.__read_parms(string_lines)

        num_rows = int(parm_dict['y-pixels'])
        num_cols = int(parm_dict['x-pixels'])
        num_pos = num_rows * num_cols
        spectra_length = int(parm_dict['z-points'])

        # num_headers = len(string_lines) - num_pos
        num_headers = 403

        # Extract the STS data from subsequent lines
        raw_data_2d = self._read_data(string_lines, num_pos, spectra_length, num_headers)

        # Generate the x / voltage / spectroscopic axis:
        volt_vec = np.linspace(-1 * max_v, 1 * max_v, spectra_length)

        h5_path = path.join(folder_path, file_name + '.h5')

        # pass on the the necessary pieces of information onto the numpy translate that will handle the creation and
        # writing to the h5 file.
        h5_path = super(AscTranslator, self).translate(h5_path, raw_data_2d, num_rows, num_cols, qty_name='Current',
                                                       data_unit='nA', spec_name='Bias', spec_unit='V',
                                                       spec_val=volt_vec, scan_height=100, scan_width=200,
                                                       spatial_unit='nm', data_type='STS', translator_name='ASC',
                                                       parms_dict=parm_dict)

        return h5_path

    def _read_data(self, string_lines, num_pos, spectra_length, num_headers):
        """
        Reads the data from lines of the data file

        Parameters
        ----------
        string_lines : list of strings
            Lines containing the data in string format, separated by tabs
        num_pos : unsigned int
            Number of pixels
        spectra_length : unsigned int
            Number of points in the spectral / voltage axis
        num_headers : unsigned int
            Number of header lines to ignore

        Returns
        -------
        raw_data_2d : 2D numpy array
            Data arranged as [position x voltage points]
        """
        raw_data_2d = np.zeros(shape=(num_pos, spectra_length), dtype=np.float32)
        for line_ind in range(num_pos):
            this_line = string_lines[num_headers + line_ind]
            string_spectrum = this_line.split('\t')[:-1]  # omitting the new line
            raw_data_2d[line_ind] = np.array(string_spectrum, dtype=np.float32)
        return raw_data_2d

    def _parse_file_path(self, input_path):
        pass

    @staticmethod
    def __read_parms(string_lines):
        """
        Returns the parameters regarding the experiment as dictionary

        Parameters
        ----------
        string_lines : list of strings
            Lines from the data file in string representation

        Returns
        -------
        parm_dict : dictionary
            Dictionary of parameters regarding the experiment
        """
        # Reading parameters stored in the first few rows of the file
        parm_dict = dict()
        for line in string_lines[3:17]:
            line = line.replace('# ', '')
            line = line.replace('\n', '')
            temp = line.split('=')
            test = temp[1].strip()
            try:
                test = float(test)
            except ValueError:
                pass
            parm_dict[temp[0].strip()] = test

        return parm_dict



"""
Created on Feb 9, 2016

@author: Chris Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals

import json
import os
import zipfile
from warnings import warn

import numpy as np
from skimage.measure import block_reduce
from skimage.util import crop

from .df_utils.io_image import unnest_parm_dicts, read_dm3
from .translator import Translator
from .utils import generate_dummy_main_parms, make_position_mat, get_spectral_slicing, \
    get_position_slicing, build_ind_val_dsets
from ..hdf_utils import getH5DsetRefs, getH5GroupRefs, calc_chunks, link_as_main
from ..io_hdf5 import ioHDF5
from ..microdata import MicroDataGroup, MicroDataset


class NDataTranslator(Translator):
    """
    Translate Pytchography data from a set of images to an HDF5 file
    """

    def __init__(self, *args, **kwargs):
        super(NDataTranslator, self).__init__(*args, **kwargs)

        self.rebin = False
        self.bin_factor = (1,1,1,1)
        self.hdf = None
        self.binning_func = self.__no_bin
        self.bin_func = None
        self.h5_main = None
        self.root_image_list = list()
        self.crop_method = 'percent'
        self.crop_ammount = None
        self.image_list_tag = None

    def translate(self, h5_path, image_path, bin_factor=None, bin_func=np.mean, start_image=0, scan_size_x=None,
                  scan_size_y=None, crop_ammount=None, crop_method='percent'):
        """
        Basic method that adds Ptychography data to existing hdf5 thisfile
        You must have already done the basic translation with BEodfTranslator

        Parameters
        ----------------
        h5_path : str
            Absolute path to where the HDF5 file should be located
        image_path : str
            Absolute path to folder holding the image files
        bin_factor : array_like of uint, optional
            Downsampling factor for each dimension.  Default is None.
        bin_func : callable, optional
            Function which will be called to calculate the return value
            of each block.  Function must implement an axis parameter,
            i.e. numpy.mean.  Ignored if bin_factor is None.  Default is
            numpy.mean.
        start_image : int, optional
            Integer denoting which image in the file path should be considered the starting
            point.  Default is 0, start with the first image on the list.
        scan_size_x : int, optional
            Number of Ronchigrams in the x direction.  Default is None, value will be determined
            from the number of images and `scan_size_y` if it is given.
        scan_size_y : int, optional
            Number of Ronchigrams in the y direction.  Default is None, value will be determined
            from the number of images and `scan_size_x` if it is given.
        crop_ammount : uint or list of uints, optional
            How much should be cropped from the original image.  Can be a single unsigned
            integer or a list of unsigned integers.  A single integer will crop the same
            ammount from all edges.  A list of two integers will crop the x-dimension by
            the first integer and the y-dimension by the second integer.  A list of 4
            integers will crop the image as [left, right, top, bottom].
        crop_method : str, optional
            Which cropping method should be used.  How much of the image is removed is
            determined by the value of `crop_ammount`.
            'percent' - A percentage of the image is removed.
            'absolute' - The specific number of pixel is removed.
        Returns
        ----------
        h5_main : h5py.Dataset
            HDF5 Dataset object that contains the flattened images

        """
        # Open the hdf5 file and delete any contents
        try:
            hdf = ioHDF5(h5_path)
            hdf.clear()
        except:
            raise

        self.hdf = hdf
        self.crop_method = crop_method
        self.crop_ammount = crop_ammount

        '''
        Get the list of all files with the .tif extension and
        the number of files in the list
        '''
        file_list = self._parse_file_path(image_path)

        '''
        Get zipfile handles for all the ndata1 files that were found in the image_path
        '''
        ziplist = [zipfile.ZipFile(zip_path, 'r') for zip_path in file_list]

        image_parm_list = self._getimageparms(file_list)

        '''
        Check if a bin_factor is given.  Set up binning objects if it is.
        '''
        if bin_factor is not None:
            self.rebin = True
            if isinstance(bin_factor, int):
                self.bin_factor = (1, 1, bin_factor, bin_factor)
            elif len(bin_factor) == 2:
                self.bin_factor = (1, 1) + bin_factor
            else:
                raise ValueError('Input parameter `bin_factor` must be a length 2 array_like or an integer.\n' +
                                 '{} was given.'.format(bin_factor))

            self.binning_func = block_reduce
            self.bin_func = bin_func

        h5_channels = self._setupH5(image_parm_list)

        self._read_data(ziplist, h5_channels)

        self.hdf.close()

        return

    def _create_root_image(self, image_path):
        """
        Create the Groups and Datasets for a single root image

        Parameters
        ----------
        image_path : str
            Path to the image file

        Returns
        -------
        None
        """
        image, image_parms = read_dm3(image_path)
        if image.ndim == 3:
            image = np.sum(image, axis=0)

        '''
        Create the Measurement and Channel Groups to hold the
        image Datasets
        '''
        root_grp = MicroDataGroup('/')

        meas_grp = MicroDataGroup('Measurement_')

        chan_grp = MicroDataGroup('Channel_')
        root_grp.addChildren([meas_grp])
        meas_grp.addChildren([chan_grp])

        '''
        Set the Measurement Group attributes
        '''
        meas_grp.attrs.update(image_parms)
        usize, vsize = image.shape
        meas_grp.attrs['image_size_u'] = usize
        meas_grp.attrs['image_size_v'] = vsize
        meas_grp.attrs['translator'] = 'OneView'
        meas_grp.attrs['num_pixels'] = image.size

        ds_rawimage = MicroDataset('Raw_Data', np.reshape(image, (-1, 1)))

        '''
        Build Spectroscopic and Position datasets for the image
        '''
        pos_mat = make_position_mat(image.shape)
        spec_mat = np.array([[0]], dtype=np.uint8)

        ds_spec_inds = MicroDataset('Spectroscopic_Indices', spec_mat)
        ds_spec_vals = MicroDataset('Spectroscopic_Values', spec_mat, dtype=np.float32)
        spec_lab = get_spectral_slicing(['Image'])
        ds_spec_inds.attrs['labels'] = spec_lab
        ds_spec_inds.attrs['units'] = ''
        ds_spec_vals.attrs['labels'] = spec_lab
        ds_spec_vals.attrs['units'] = ''

        ds_pos_inds = MicroDataset('Position_Indices', pos_mat)
        ds_pos_vals = MicroDataset('Position_Values', pos_mat, dtype=np.float32)

        pos_lab = get_position_slicing(['X', 'Y'])
        ds_pos_inds.attrs['labels'] = pos_lab
        ds_pos_inds.attrs['units'] = ['pixel', 'pixel']
        ds_pos_vals.attrs['labels'] = pos_lab
        ds_pos_vals.attrs['units'] = ['pixel', 'pixel']

        chan_grp.addChildren([ds_rawimage, ds_spec_inds, ds_spec_vals,
                              ds_pos_inds, ds_pos_vals])

        '''
        Write the data to file and get the handle for the image dataset
        '''
        image_refs = self.hdf.writeData(root_grp)

        h5_image = getH5DsetRefs(['Raw_Data'], image_refs)[0]

        '''
        Link references to raw
        '''
        aux_ds_names = ['Position_Indices', 'Position_Values', 'Spectroscopic_Indices', 'Spectroscopic_Values']
        link_as_main(h5_image, *getH5DsetRefs(aux_ds_names, image_refs))

        self.root_image_list.append(h5_image)

    def _read_data(self, zip_list, h5_channels):
        """
        Iterates over the images in `file_list`, reading each image and downsampling if
        reqeusted, and writes the flattened image to file.  Also builds the Mean_Ronchigram
        and the Spectroscopic_Mean datasets at the same time.

        Parameters
        ----------
        zip_list : list of str
            List of all files in `image_path` that will be read
        h5_main : h5py.Dataset
            Dataset which will hold the Ronchigrams
        h5_mean_spec : h5py.Dataset
            Dataset which will hold the Spectroscopic Mean
        h5_ronch : h5py.Dataset
            Dataset which will hold the Mean Ronchigram
        image_path : str
            Absolute file path to the directory which hold the images

        Returns
        -------
        None
        """
        h5_main_list = list()
        '''
        For each file, we must read the data then create the neccessary datasets, add them to the channel, and
        write it all to file
        '''
        for ifile, (this_file, this_channel) in enumerate(zip(zip_list, h5_channels)):
            '''
            Extract the data file from the zip archive and read it into an array
            '''
            tmp_path = this_file.extract('data.npy')
            this_data = np.load(tmp_path)
            os.remove(tmp_path)

            '''
            Find the shape of the data, then calculate the final dimensions based on the crop and
            downsampling parameters
            '''
            while this_data.ndim < 4:
                this_data = np.expand_dims(this_data, 0)

            this_data = self.crop_ronc(this_data)
            scan_size_x, scan_size_y, usize, vsize = this_data.shape

            usize = int(round(1.0*usize / self.bin_factor[-2]))
            vsize = int(round(1.0*vsize / self.bin_factor[-1]))

            num_images = scan_size_x*scan_size_y
            num_pixels = usize*vsize

            '''
            Write these attributes to the Measurement group
            '''
            new_attrs = {'image_size_u': usize,
                         'image_size_v': vsize,
                         'scan_size_x': scan_size_x,
                         'scan_size_y': scan_size_y}
            this_channel.parent.attrs.update(new_attrs)

            # Get the Position and Spectroscopic Datasets
            ds_spec_ind, ds_spec_vals = build_ind_val_dsets((usize, vsize), is_spectral=True,
                                                            labels=['U', 'V'], units=['pixel', 'pixel'])
            ds_pos_ind, ds_pos_val = build_ind_val_dsets([scan_size_x, scan_size_y], is_spectral=False,
                                                         labels=['X', 'Y'], units=['pixel', 'pixel'])

            ds_chunking = calc_chunks([num_images, num_pixels],
                                      np.float32(0).itemsize,
                                      unit_chunks=(1, num_pixels))

            # Allocate space for Main_Data and Pixel averaged Data
            ds_main_data = MicroDataset('Raw_Data', data=[], maxshape=(num_images, num_pixels),
                                        chunking=ds_chunking, dtype=np.float32, compression='gzip')
            ds_mean_ronch_data = MicroDataset('Mean_Ronchigram',
                                              data=np.zeros(num_pixels, dtype=np.float32),
                                              dtype=np.float32)
            ds_mean_spec_data = MicroDataset('Spectroscopic_Mean',
                                             data=np.zeros(num_images, dtype=np.float32),
                                             dtype=np.float32)

            # Add datasets as children of Measurement_000 data group
            ds_channel = MicroDataGroup(this_channel.name)
            ds_channel.addChildren([ds_main_data, ds_spec_ind, ds_spec_vals, ds_pos_ind,
                                    ds_pos_val, ds_mean_ronch_data, ds_mean_spec_data])

            h5_refs = self.hdf.writeData(ds_channel)
            h5_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
            h5_ronch = getH5DsetRefs(['Mean_Ronchigram'], h5_refs)[0]
            h5_mean_spec = getH5DsetRefs(['Spectroscopic_Mean'], h5_refs)[0]

            aux_ds_names = ['Position_Indices',
                            'Position_Values',
                            'Spectroscopic_Indices',
                            'Spectroscopic_Values']

            link_as_main(h5_main, *getH5DsetRefs(aux_ds_names, h5_refs))

            mean_ronch = np.zeros(h5_ronch.shape, dtype=np.float32)

            this_data = self.binning_func(this_data, self.bin_factor, self.bin_func).reshape(h5_main.shape)

            h5_main[:, :] = this_data

            h5_mean_spec[:] = np.mean(this_data, axis=1)

            h5_ronch[:] = np.mean(this_data, axis=0)

            self.hdf.flush()

            h5_main_list.append(h5_main)

        self.hdf.flush()

    def crop_ronc(self, ronc):
        """
        Crop the input Ronchigram by the specified ammount using the specified method.

        Parameters
        ----------
        ronc : numpy.array
            Input image to be cropped.

        Returns
        -------
        cropped_ronc : numpy.array
            Cropped image
        """

        if self.crop_ammount is None:
            return ronc

        crop_ammount = self.crop_ammount
        crop_method = self.crop_method

        if crop_method == 'percent':
            crop_ammount = np.round(np.atleast_2d(crop_ammount)/100.0*ronc.shape)
            crop_ammount = tuple([tuple(row) for row in crop_ammount.astype(np.uint32)])
        elif crop_method == 'absolute':
            if isinstance(crop_ammount, int):
                crop_ammount = ((crop_ammount,), (crop_ammount,))
            elif len(crop_ammount) == 2:
                crop_ammount = ((crop_ammount[0],), (crop_ammount[1],))
            elif len(crop_ammount) == 4:
                crop_ammount = ((crop_ammount[0], crop_ammount[1]), (crop_ammount[2], crop_ammount[3]))
            else:
                raise ValueError('The crop_ammount should be an integer or list of 2 or 4 integers.')
        else:
            raise ValueError('Allowed values of crop_method are percent and absolute.')

        crop_ammount = ((0,), (0,))+crop_ammount

        cropped_ronc = crop(ronc, crop_ammount)

        if any([dim == 0 for dim in cropped_ronc.shape]):
            warn("Requested crop ammount is greater than the image size.  No cropping will be done.")
            return ronc
        return cropped_ronc

    def downSampRoncVec(self, ronch_vec, binning_factor):
        """
        Downsample the image by taking the mean over nearby values

        Parameters
        ----------
        ronch_vec : ndarray
            Image data
        binning_factor : int
            factor to reduce the size of the image by

        Returns
        -------
        ronc_mat3_mean : ndarray
            Flattened downsampled image
        """
        ccd_pix = int(np.sqrt(ronch_vec.size))
        ronc_mat = ronch_vec.reshape(ccd_pix, ccd_pix)
        ronc_mat2 = ronc_mat.reshape(ccd_pix, ccd_pix / binning_factor, binning_factor)
        ronc_mat2_mean = ronc_mat2.mean(2)  # take the mean along the 3rd dimension
        ronc_mat3 = ronc_mat2_mean.reshape(ccd_pix / binning_factor, binning_factor, -1)
        ronc_mat3_mean = ronc_mat3.mean(1)

        return ronc_mat3_mean.reshape(-1)

    @staticmethod
    def _parse_file_path(image_folder):
        """
        Returns a list of all files in the directory given by path

        Parameters
        ---------------
        path : string / unicode
            absolute path to directory containing files

        Returns
        ----------
        file_list : list of strings
            names of all files in directory located at path
        """
        file_list = list()
        allowed_image_types = ['.ndata1']
        for root, dirs, files in os.walk(image_folder):
            for thisfile in files:
                _, ext = os.path.splitext(thisfile)
                if ext not in allowed_image_types:
                    continue
                else:
                    file_list.append(os.path.join(root, thisfile))
        return file_list

    @staticmethod
    def _getimageparms(file_list):
        """
        Returns the image parameters for each file in the `file_list`

        Parameters
        ------------
        file_list : list of zipfile.ZipFile
            List of zipfile objects

        Returns
        -----------
        parm_list : list of dict
            List of image parameters from the files in `file_list`
        """
        parm_list = list()

        for zpath in file_list:
            zfile = zipfile.ZipFile(zpath, 'r')
            tmp_path = zfile.extract('metadata.json')
            metafile = open(tmp_path, 'r')
            metastring = metafile.read()
            parm_list.append(unnest_parm_dicts(json.loads(metastring)))
            metafile.close()

            os.remove(tmp_path)

        return parm_list

    def _setupH5(self, image_parms):
        """
        Setup the HDF5 file in which to store the data
        Due to the structure of the ndata format, we can only create the Measurement and Channel groups here

        Parameters
        ----------
        image_parms : dict
            Dictionary of parameters

        Returns
        -------
        h5_main : h5py.Dataset
            HDF5 Dataset that the images will be written into
        h5_mean_spec : h5py.Dataset
            HDF5 Dataset that the mean over all positions will be written
            into
        h5_ronch : h5py.Dataset
            HDF5 Dateset that the mean over all Spectroscopic steps will be
            written into
        """
        root_parms = generate_dummy_main_parms()
        root_parms['data_type'] = 'PtychographyData'

        # Create the hdf5 data Group
        root_grp = MicroDataGroup('/')
        root_grp.attrs = root_parms

        dg_channels = list()
        for meas_parms in image_parms:
            # Create new measurement group for each set of parameters
            meas_grp = MicroDataGroup('Measurement_')
            # Write the parameters as attributes of the group
            for key in meas_parms.keys():
                meas_grp.attrs[key] = meas_parms[key]
            chan_grp = MicroDataGroup('Channel_000')

            meas_grp.addChildren([chan_grp])
            root_grp.addChildren([meas_grp])



        # Write the groups to file
        h5_refs = self.hdf.writeData(root_grp)

        h5_channels = getH5GroupRefs('Channel_000', h5_refs)

        self.hdf.flush()

        return h5_channels

    @staticmethod
    def __no_bin(image, *args, **kwargs):
        """
        Does absolutely nothing to the image.  Exists so that we can have
        a bin function to call whether we actually rebin the image or not.

        Parameters
        ----------
        image : ndarray
            Image
        args:
            Argument list
        kwargs:
            Keyword argument list

        Returns
        -------
        image : ndarray
            The input image
        """
        return image



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 07 15:21:46 2015

@author: Suhas Somnath
"""
from __future__ import division, print_function, absolute_import, unicode_literals

from os import path, listdir, remove
from warnings import warn

import numpy as np
from scipy.io.matlab import loadmat  # To load parameters stored in Matlab .mat file

from .df_utils.be_utils import parmsToDict
from .translator import Translator
from .utils import generate_dummy_main_parms, build_ind_val_dsets
from ..hdf_utils import getH5DsetRefs, linkRefs
from ..io_hdf5 import ioHDF5
from ..microdata import MicroDataGroup, MicroDataset


class GLineTranslator(Translator):
    """
    Translated G-mode line (bigtimedata.dat) files from actual BE line experiments to HDF5
    """
    
    def translate(self, file_path):
        """
        The main function that translates the provided file into a .h5 file
        
        Parameters
        ----------
        file_path : String / unicode
            Absolute path of any file in the directory

        Returns
        -------
        h5_path : String / unicode
            Absolute path of the h5 file
        """
        # Figure out the basename of the data:
        (basename, parm_paths, data_paths) = self._parse_file_path(file_path)
        
        (folder_path, unused) = path.split(file_path)
        h5_path = path.join(folder_path, basename+'.h5')
        
        if path.exists(h5_path):
            remove(h5_path)
        
        # Load parameters from .mat file - 'BE_wave', 'FFT_BE_wave', 'total_cols', 'total_rows'
        matread = loadmat(parm_paths['parm_mat'], variable_names=['BE_wave', 'FFT_BE_wave', 'total_cols', 'total_rows'])
        be_wave = np.float32(np.squeeze(matread['BE_wave']))

        # Need to take the complex conjugate if reading from a .mat file
        # FFT_BE_wave = np.conjugate(np.complex64(np.squeeze(matread['FFT_BE_wave'])))
        
        num_cols = int(matread['total_cols'][0][0])
        expected_rows = int(matread['total_rows'][0][0])
        self.points_per_pixel = len(be_wave)
        
        # Load parameters from .txt file - 'BE_center_frequency_[Hz]', 'IO rate'
        is_beps, parm_dict = parmsToDict(parm_paths['parm_txt'])
        
        # Get file byte size:
        # For now, assume that bigtime_00 always exists and is the main file
        file_size = path.getsize(data_paths[0])
        
        # Calculate actual number of lines since the first few lines may not be saved
        self.num_rows = 1.0 * file_size / (4 * self.points_per_pixel * num_cols)
        if self.num_rows % 1:
            warn('Error - File has incomplete rows')
            return None
        else:
            self.num_rows = int(self.num_rows)

        samp_rate = parm_dict['IO_rate_[Hz]']
        ex_freq_nominal = parm_dict['BE_center_frequency_[Hz]']

        # method 1 for calculating the correct excitation frequency:
        pixel_duration = 1.0 * self.points_per_pixel / samp_rate
        num_periods = pixel_duration * ex_freq_nominal
        ex_freq_correct = 1 / (pixel_duration / np.floor(num_periods))

        # method 2 for calculating the exact excitation frequency:
        """
        fft_ex_wfm = np.abs(np.fft.fftshift(np.fft.fft(be_wave)))
        w_vec = np.linspace(-0.5 * samp_rate, 0.5 * samp_rate - 1.0*samp_rate / self.points_per_pixel,
                            self.points_per_pixel)
        hot_bins = np.squeeze(np.argwhere(fft_ex_wfm > 1E+3))
        ex_freq_correct = w_vec[hot_bins[-1]]
        """

        # correcting the excitation frequency - will be VERY useful during analysis and filtering
        parm_dict['BE_center_frequency_[Hz]'] = ex_freq_correct

        # Some very basic information that can help the processing crew
        parm_dict['num_bins'] = self.points_per_pixel
        parm_dict['grid_num_rows'] = self.num_rows
        parm_dict['data_type'] = 'G_mode_line'
            
        if self.num_rows != expected_rows:
            print('Note: {} of {} lines found in data file'.format(self.num_rows, expected_rows))
        
        # Calculate number of points to read per line:
        self.__bytes_per_row__ = int(file_size/self.num_rows)

        # First finish writing all global parameters, create the file too:
        meas_grp = MicroDataGroup('Measurement_000')
        meas_grp.attrs = parm_dict

        spm_data = MicroDataGroup('')
        global_parms = generate_dummy_main_parms()
        global_parms['data_type'] = 'G_mode_line'
        global_parms['translator'] = 'G_mode_line'
        spm_data.attrs = global_parms
        spm_data.addChildren([meas_grp])
        
        hdf = ioHDF5(h5_path)
        # hdf.clear()
        hdf.writeData(spm_data)
        
        # Now that the file has been created, go over each raw data file:
        # 1. write all ancillary data. Link data. 2. Write main data sequentially
                
        """ We only allocate the space for the main data here.
        This does NOT change with each file. The data written to it does.
        The auxiliary datasets will not change with each raw data file since
        only one excitation waveform is used"""
        ds_main_data = MicroDataset('Raw_Data', data=[], 
                                    maxshape=(self.num_rows, self.points_per_pixel * num_cols),
                                    chunking=(1, self.points_per_pixel), dtype=np.float16)
        ds_main_data.attrs['quantity'] = ['Deflection']
        ds_main_data.attrs['units'] = ['V']

        ds_pos_ind, ds_pos_val = build_ind_val_dsets([self.num_rows], is_spectral=False,
                                                     labels=['Y'], units=['m'])
        ds_spec_inds, ds_spec_vals = build_ind_val_dsets([self.points_per_pixel * num_cols], is_spectral=True,
                                                         labels=['Excitation'], units=['V'])
        ds_spec_vals.data = np.atleast_2d(np.tile(np.float32(be_wave), num_cols))  # Override the default waveform
        
        aux_ds_names = ['Position_Indices', 'Position_Values',
                        'Spectroscopic_Indices', 'Spectroscopic_Values']
        
        for f_index in data_paths.keys():
            
            chan_grp = MicroDataGroup('{:s}{:03d}'.format('Channel_', f_index), '/Measurement_000/')
            chan_grp.addChildren([ds_main_data, ds_pos_ind, ds_pos_val, ds_spec_inds, ds_spec_vals])
            
            # print('Writing following tree to file:')
            # chan_grp.showTree()
            h5_refs = hdf.writeData(chan_grp)
            
            h5_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]  # We know there is exactly one main data
            
            # Reference linking can certainly take place even before the datasets have reached their final size         
            linkRefs(h5_main, getH5DsetRefs(aux_ds_names, h5_refs))
            
            # Now transfer scan data in the dat file to the h5 file:
            self._read_data(data_paths[f_index], h5_main)
            
        hdf.close()
        print('G-Line translation complete!')

        return h5_path

    @staticmethod
    def _parse_file_path(data_filepath):
        """
        Goes through the file directory and figures out the basename and the 
        parameter (text and .mat), data file paths (for each analog input channel)
        
        Parameters
        -----------------
        data_filepath : string / unicode
            absolute path of any file in the data folder
        
        Returns
        ----------------
        basename : string / unicode
            base name of the experiment\n
        parm_paths : dictionary
            paths for the text and .mat parameter files\n
            parm_text : absolute file path of the parameter text file\n
            parm_mat : absolute file path of the parameter .mat file
        data_paths : dictionary of the paths for the big-time data files.
            key : index of the analog input that generated the data file\n
            value : absolute file path of the data file
        """
        # Return (basename, parameter text path)
        (folder_path, basename) = path.split(data_filepath)
        (upper_folder, basename) = path.split(folder_path)
        
        # There may be one or two bigdata files. May need both paths
        parm_paths = dict()
        data_paths = dict()
        targ_str = 'bigtime_0'
        for filenames in listdir(folder_path):
            ind = filenames.find(targ_str)
            if ind > 0 and filenames.endswith('.dat'):
                data_paths[int(filenames[ind+len(targ_str)])] = path.join(folder_path, filenames)
        
            if filenames.endswith('.txt') and filenames.find('parm') > 0:
                parm_paths['parm_txt'] = path.join(folder_path, filenames)
                
            if filenames.endswith('_all.mat'):
                parm_paths['parm_mat'] = path.join(folder_path, filenames)
                
        return basename, parm_paths, data_paths

    def _read_data(self, filepath, h5_dset):
        """
        Reads the .dat file and populates the .h5 dataset

        Parameters
        ---------
        filepath : String / unicode
            absolute path of the data file for a particular analog input channel
        h5_dset : HDF5 dataset reference
            Reference to the target Raw_Data dataset

        Returns
        ---------
        None
        """
        # Create data matrix - Only need 16 bit floats (time) 
                
        # Read line by line and write to h5                 
        with open(filepath, 'rb') as file_handl:
            for row_indx in range(self.num_rows):
                
                if row_indx % 10 == 0:
                    print('Reading line {} of {}'.format(row_indx, self.num_rows))
                
                file_handl.seek(row_indx*self.__bytes_per_row__, 0)
                data_vec = np.fromstring(file_handl.read(self.__bytes_per_row__), dtype='f')
                h5_dset[row_indx] = np.float16(data_vec)
                h5_dset.file.flush()
        
        print('Finished reading file: {}!'.format(filepath))



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 29 16:04:34 2016

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import, unicode_literals

from os import path, remove # File Path formatting

import h5py
import numpy as np  # For array operations
from scipy.io.matlab import loadmat  # To load parameters stored in Matlab .mat file

from .translator import Translator  # Because this class extends the abstract Translator class
from .utils import make_position_mat, get_position_slicing, generate_dummy_main_parms
from ..hdf_utils import getH5DsetRefs, linkRefs
from ..io_hdf5 import ioHDF5  # Now the translator is responsible for writing the data.
from ..microdata import MicroDataGroup, MicroDataset  # The building blocks for defining heirarchical storage in the H5 file


class SporcTranslator(Translator):
    """
    Translates G-mode SPORC datasets from .mat files to .h5
    """
    
    def translate(self, parm_path):
        """
        Basic method that translates .mat data files to a single .h5 file
        
        Parameters
        ------------
        parm_path : string / unicode
            Absolute file path of the parameters .mat file. 
            
        Returns
        ----------
        h5_path : string / unicode
            Absolute path of the translated h5 file
        """
		
        (folder_path, file_name) = path.split(parm_path)
        (file_name, base_name) = path.split(folder_path)
        h5_path = path.join(folder_path,base_name+'.h5')
        
        # Read parameters
        print('reading parameter files')
        parm_dict, excit_wfm, spec_ind_mat = self.__readparms(parm_path)
        parm_dict['data_type'] = 'SPORC'
        
        num_rows = parm_dict['grid_num_rows']
        num_cols = parm_dict['grid_num_cols'] 
        num_pix = num_rows*num_cols        
        
        pos_mat = make_position_mat([num_cols, num_rows])
        pos_slices = get_position_slicing(['X', 'Y'], num_pix)
        
        # new data format
        spec_ind_mat = np.transpose(np.float32(spec_ind_mat))
        
        # Now start creating datasets and populating:
        ds_pos_ind = MicroDataset('Position_Indices', np.uint32(pos_mat))
        ds_pos_ind.attrs['labels'] = pos_slices
        ds_pos_val = MicroDataset('Position_Values', np.float32(pos_mat))
        ds_pos_val.attrs['labels'] = pos_slices
        ds_pos_val.attrs['units'] = ['um','um']
                
        spec_ind_labels = ['x index', 'y index', 'loop index', 'repetition index', 'slope index']
        spec_ind_dict = dict()
        for col_ind, col_name in enumerate(spec_ind_labels):
            spec_ind_dict[col_name] = (slice(col_ind,col_ind+1), slice(None))
        ds_spec_inds = MicroDataset('Spectroscopic_Indices', np.uint32(spec_ind_mat))
        ds_spec_inds.attrs['labels'] = spec_ind_dict
        ds_spec_vals = MicroDataset('Spectroscopic_Values', spec_ind_mat)
        ds_spec_vals.attrs['labels'] = spec_ind_dict
        ds_spec_vals.attrs['units'] = ['V','V','','','']

        ds_excit_wfm = MicroDataset('Excitation_Waveform', np.float32(excit_wfm))
        
        ds_raw_data = MicroDataset('Raw_Data', data=[], 
                                  maxshape=(num_pix,len(excit_wfm)), 
                                dtype=np.float16, chunking=(1,len(excit_wfm)), 
                                compression='gzip')
        
        # technically should change the date, etc.
         
        chan_grp = MicroDataGroup('Channel_000')
        chan_grp.attrs = parm_dict        
        chan_grp.addChildren([ds_pos_ind, ds_pos_val, ds_spec_inds, ds_spec_vals,
                              ds_excit_wfm, ds_raw_data])
                
        global_parms = generate_dummy_main_parms()
        global_parms['grid_size_x'] = parm_dict['grid_num_cols'];
        global_parms['grid_size_y'] = parm_dict['grid_num_rows'];
        # assuming that the experiment was completed:        
        global_parms['current_position_x'] = parm_dict['grid_num_cols']-1;
        global_parms['current_position_y'] = parm_dict['grid_num_rows']-1;
        global_parms['data_type'] = parm_dict['data_type'] 
        global_parms['translator'] = 'SPORC'
        
        meas_grp = MicroDataGroup('Measurement_000')
        meas_grp.addChildren([chan_grp])
        spm_data = MicroDataGroup('')
        spm_data.attrs = global_parms
        spm_data.addChildren([meas_grp])
        
        if path.exists(h5_path):
            remove(h5_path)
        
        # Write everything except for the main data.
        hdf = ioHDF5(h5_path)
        
        h5_refs = hdf.writeData(spm_data)
                    
        h5_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
            
        #Now doing linkrefs:
        aux_ds_names = ['Excitation_Waveform', 'Position_Indices','Position_Values',
                     'Spectroscopic_Indices','Spectroscopic_Values']
        linkRefs(h5_main, getH5DsetRefs(aux_ds_names, h5_refs))
        
        print('reading raw data now...')
        
        # Now read the raw data files:
        pos_ind = 0
        for row_ind in range(1, num_rows+1):
            for col_ind in range(1, num_cols+1):
                file_path = path.join(folder_path,'result_r'+str(row_ind)+'_c'+str(col_ind)+'.mat')
                #print('Working on row {} col {}'.format(row_ind,col_ind))
                if path.exists(file_path):
                    # Load data file
                    pix_data = loadmat(file_path, squeeze_me=True)
                    # Take the inverse FFT on 1st dimension
                    pix_vec = np.fft.ifft(np.fft.ifftshift(pix_data['data']))
                    # Verified with Matlab - no conjugate required here.
                    h5_main[pos_ind,:] = np.float16(np.real(pix_vec))
                    hdf.flush() # flush from memory!
                else:
                    print('File for row {} col {} not found'.format(row_ind, col_ind))
                pos_ind +=1
                if (100.0*(pos_ind)/num_pix)%10 == 0:
                    print('Finished reading {} % of data'.format(int(100*pos_ind/num_pix)))

        hdf.close()
        
        return h5_path
        
        
    @staticmethod
    def __readparms(parm_path):
        """
        Copies experimental parameters from the .mat file to a dictionary

        Parameters
        ------------
        parm_path : string / unicode
            Absolute path of the parameters file

        Returns
        --------
        (parm_dict, excit_wfm, spec_ind_mat) : tuple

        parm_dict : dictionary
            Dictionary containing all relevant parameters
        excit_wfm : 1d numpy float array
            Excitation waveform
        spec_ind_mat : 2D numpy flaot array
            Spectroscopic indicies matrix
        """

        parm_data = loadmat(parm_path, squeeze_me=True, struct_as_record=True)
        parm_dict = dict()

        IO_parms = parm_data['IOparms']
        parm_dict['IO_samp_rate_[Hz]'] = np.int32(IO_parms['sampRate'].item())
        parm_dict['IO_down_samp_rate_[Hz]'] = np.int32(IO_parms['downSampRate'].item())
        parm_dict['IO_AO0_amp'] = np.int32(IO_parms['AO0_amp'].item())
        parm_dict['IO_AI_chans'] = np.int32(parm_data['aiChans'])

        parm_dict['grid_num_rows'] = parm_data['numrows']
        parm_dict['grid_num_cols'] = parm_data['numcols']

        sporc_parms = parm_data['sporcParms']
        parm_dict['SPORC_V_max_[V]'] = np.float32(sporc_parms['V_max'].item())
        parm_dict['SPORC_N_steps'] = np.int32(sporc_parms['N_steps'].item())
        parm_dict['SPORC_N_reps'] = np.int32(sporc_parms['N_reps'].item())
        parm_dict['SPORC_t_max_[sec]'] = np.float32(sporc_parms['t_max'])
        parm_dict['SPORC_f_cutoff_[Hz]'] = np.int32(sporc_parms['f_cutoff'])
        parm_dict['SPORC_f_rolloff_[Hz]'] = np.int32(sporc_parms['f_rolloff'])

        if 'FORC_vec' in parm_data.keys() and 'ind_vecs' in parm_data.keys():
            excit_wfm = np.squeeze(np.float32(parm_data['FORC_vec']))
            spec_ind_mat = np.transpose(np.float32(parm_data['ind_vecs']))
        else:
            # Look for a second parms file that contains these vectors:
            fold, basename = path.split(parm_path)
            second_path = path.join(fold,'SPORC_wave.mat')
            h5_sporc_parms = h5py.File(second_path,'r') # Use this for v7.3 and beyond.
            excit_wfm = np.squeeze(h5_sporc_parms['FORC_vec'].value)
            spec_ind_mat = np.float32(h5_sporc_parms['ind_vecs'].value)
            h5_sporc_parms.close()

        return parm_dict, excit_wfm, spec_ind_mat


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 16 10:42:03 2015

@author: Suhas Somnath, Numan Laanait

The MicroData classes

"""

from __future__ import division, print_function, absolute_import, unicode_literals
import socket
from warnings import warn

from .io_utils import getTimeStamp


class MicroData(object):
    """
    Generic class that is extended by the MicroDataGroup and MicroDataset objects
    """
    
    def __init__(self, name, parent):
        '''
        Parameters
        ----------
        name : String
            Name of the data group / dataset object
        parent : String
            HDF5 path to the parent of this object. Typically used when
            appending to an existing HDF5 file
        '''
        self.name = name
        self.attrs = dict()
        self.parent = parent
        self.indexed = False

class MicroDataGroup(MicroData):
    """
    Holds data that will be converted to a h5.Group by io.ioHDF5
    Note that it can also hold information (e.g. attributes) of an h5.File.
    This is consistent with class hierarchy of HDF5, i.e. h5.File extends h5.Group.
    """
    def __init__(self, name, parent='/'):
        """
        Parameters
        ----------
        name : String
            Name of the data group
        parent : (Optional) String
            HDF5 path to the parent of this object. Typically used when
            appending to an existing HDF5 file
            Default value assumes that this group sits at the root of the file
        """
        super(MicroDataGroup, self).__init__(name, parent)
        self.children = list()
        self.attrs['machine_id'] = socket.getfqdn()
        self.attrs['timestamp'] = getTimeStamp()

        if name != '':
            self.indexed = self.name[-1] == '_'
        
        pass
    
    def addChildren(self, children): 
        '''
        Adds Children to the class to make a tree structure.
        
        Parameters
        ----------
        children : list of MicroData objects
            Children can be a mixture of groups and datasets
            
        Returns
        -------
        None
        '''
        for child in children:
            if isinstance(child, MicroData):
                child.parent = self.parent+self.name
                self.children.append(child)
            else:
                warn('Children must be of type MicroData.')

    def __str__(self):
        self.showTree()
            
    def showTree(self):
        """
        Return the tree structure given by MicroDataGroup.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        def __tree(child, parent):
            print(parent+'/'+child.name)
            if isinstance(child, MicroDataGroup):
                for ch in child.children:
                    __tree(ch, parent+'/'+child.name)
        
        # print(self.parent+self.name)
        for child in self.children:
            __tree(child, self.parent+self.name)
                    
    
class MicroDataset(MicroData):
    """
    Holds data (i.e. numpy.ndarray) as well as instructions on writing, attributes, etc...
    This gets converted to a h5.Dataset by io.ioHDF5.\n    

    Region references need to be specified using the 'labels' attribute. See example below    
    """
    
    def __init__(self, name, data, dtype=None, compression=None, chunking=None, parent=None, resizable=False,
                 maxshape=None):
        """
        Parameters
        ----------
        name : String
            Name of the dataset
        data : Object
            See supported objects in h5py
        dtype : datatype, 
            typically a datatype of a numpy array =None
        compression : (Optional) String
            See h5py compression. Leave as 'gzip' as a default mode of compression
        chunking : (Optional) tuple of ints
            Chunking in each dimension of the dataset. If not provided, 
            default chunking is used by h5py when writing this dataset
        parent : (Optional) String
                HDF5 path to the parent of this object. This value is overwritten
                when this dataset is made the child of a datagroup.
        resizable : Boolean (Optional. default = False)
            Whether or not this dataset is expected to be resizeable.
            Note - if the dataset is resizable the specified maxsize will be ignored. 
        maxshape : (Optional) tuple of ints
            Maximum size in each axis this dataset is expected to be
            if this parameter is provided, io will ONLY allocate space. 
            Make sure to specify the dtype appropriately. The provided data will be ignored
            
        Examples
        --------   
        1. Small auxillary datasets :
        
        >>> ds_ex_efm = MicroDataset('Excitation_Waveform', np.arange(10))
        
        2. Datasets with slicing information :
        
        >>> ds_spec_inds = MicroDataset('Spectroscopic_Indices', spec_ind_mat)
            ds_spec_inds.attrs['labels'] = {'Time Index':(slice(0,1), slice(None))}
            
        3. Initializing large primary datasets of known sizes :
        
        >>> ds_raw_data = MicroDataset('Raw_Data', data=[], maxshape=(1024,16384), dtype=np.float16, chunking=(1,16384), compression='gzip')
                    
        4. Intializing large datasets whose size is unknown in one or more dimensions:
        
        >>> ds_raw_data = MicroDataset('Raw_Data', np.zeros(shape=(1,16384), dtype=np.complex64), chunking=(1,16384), resizable=True,compression='gzip')
        """

        def _make_iterable(item):
            if item is not None:
                if type(item) not in [list, tuple]:  # another (inelegant) way of asking if this object is iterable
                    item = tuple([item])
            return item

        super(MicroDataset, self).__init__(name, parent)
        self.data = data
        self.dtype = dtype
        self.compression = compression
        self.chunking = _make_iterable(chunking)
        self.resizable = resizable
        self.maxshape = _make_iterable(maxshape)
        if resizable is True:
            self.maxshape = None  # Overridden
            self.shape = None
        elif maxshape is not None:
            self.shape = self.maxshape
        else:
            self.shape = self.data.shape

    def __getitem__(self, item):
        return self.data[item]


# -*- coding: utf-8 -*-
"""
Created on Fri Jan 27 17:58:35 2017

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import numpy as np  # For array operations

from .translator import Translator
from .utils import build_ind_val_dsets
from ..hdf_utils import calc_chunks
from ..microdata import MicroDataset  # building blocks for defining hierarchical storage in the H5 file


class NumpyTranslator(Translator):
    """
    Writes a numpy array to .h5
    """

    def _read_data(self):
        pass

    def _parse_file_path(self, input_path):
        pass

    def translate(self, h5_path, main_data, num_rows, num_cols, qty_name='Unknown', data_unit='a. u.',
                  spec_name='Spectroscopic_Variable', spec_val=None, spec_unit='a. u.', data_type='generic',
                  translator_name='numpy', scan_height=None, scan_width=None, spatial_unit='m', parms_dict={}):
        """
        The main function that translates the provided data into a .h5 file

        Parameters
        ----------
        h5_path
        main_data
        num_rows
        num_cols
        qty_name
        data_unit
        spec_name
        spec_val
        spec_unit
        data_type
        translator_name
        scan_height
        scan_width
        spatial_unit
        parms_dict

        Returns
        -------
        h5_path : string / unicode
            Absolute path of the translated h5 file
        """
        if main_data.ndim != 2:
            raise ValueError('Main dataset must be a 2-dimensional array arranged as [positions x spectra]')

        spectra_length = main_data.shape[1]

        ds_main = MicroDataset('Raw_Data', data=main_data, dtype=np.float32, compression='gzip',
                               chunking=calc_chunks(main_data.shape, np.float32(0).itemsize,
                                                    unit_chunks=(1, spectra_length)))
        ds_main.attrs = {'quantity': qty_name, 'units': data_unit}

        pos_steps = None
        if scan_width is not None and scan_height is not None:
            pos_steps = [1.0 * scan_height / num_rows, 1.0 * scan_width / num_cols]

        ds_pos_ind, ds_pos_val = build_ind_val_dsets([num_rows, num_cols], is_spectral=False, steps=pos_steps,
                                                     labels=['Y', 'X'], units=[spatial_unit, spatial_unit],
                                                     verbose=False)
        ds_spec_inds, ds_spec_vals = build_ind_val_dsets([spectra_length], is_spectral=True,
                                                         labels=[spec_name], units=[spec_unit], verbose=False)
        if spec_val is not None:
            if type(spec_val) in [list, np.ndarray]:
                ds_spec_vals.data = np.float32(np.atleast_2d(spec_val))

        parms_dict.update({'translator': 'NumpyTranslator'})

        return super(NumpyTranslator, self).simple_write(h5_path, data_type, translator_name, ds_main,
                                                         [ds_pos_ind, ds_pos_val, ds_spec_inds, ds_spec_vals],
                                                         parm_dict=parms_dict)
